import copy
import csv
from typing import Optional, Iterable

from ..common.reader import zip_columns_with_values
from ..common.multiheader_csv.config import MultiheaderConfig
from ..common.multiheader_csv.core import MultiheaderCsvCore

class MHdrCsvWriter:
    _multiheader_core: MultiheaderCsvCore

    def __init__(self, f, headers: dict[str, list[str]], restval="", extrasaction="raise",
                 dialect="excel", multiheader_config: Optional[MultiheaderConfig] = None,
                 *args, **kwds):
        self.restval = restval          # for writing short dicts
        if extrasaction.lower() not in ("raise", "ignore"):
            raise ValueError("extrasaction (%s) must be 'raise' or 'ignore'"
                             % extrasaction)
        self.extrasaction = extrasaction
        self._multiheader_core = MultiheaderCsvCore(multiheader_config)
        self._multiheader_core.set_headers(headers, writable=True)
        self.writer = csv.writer(f, dialect, *args, **kwds)

    @property
    def headers(self) -> dict[str, list[str]]:
        return self._multiheader_core.headers

    @headers.setter
    def headers(self, headers: dict[str, list[str]]):
        self._multiheader_core.set_headers(headers, writable=True)

    def writeheaders(self):
        return self.writer.writerows(self._multiheader_core.generate_headers_definitions())

    def writerow(self, rowdict: dict):
        return self.writer.writerow(self._multiheader_core.generate_writable_line(rowdict))

    def writerows(self, rowdicts: Iterable[dict]):
        return self.writer.writerows(map(self._multiheader_core.generate_writable_line, rowdicts))


from tabbyset.utils.flex_table import FlexTableValue
from enum import Enum
from typing import Optional


class ConsoleColor(Enum):
    RED = "31"
    GREEN = "32"
    YELLOW = "33"
    BLUE = "34"
    MAGENTA = "35"
    CYAN = "36"
    WHITE = "37"
    GRAY = "90"


class ColoredString:
    def __init__(self, value: str, color: ConsoleColor):
        if not isinstance(value, str):
            raise TypeError(f"Value must be a string ({value})")
        if not isinstance(color, ConsoleColor):
            raise TypeError(f"Color must be a ConsoleColor ({color})")
        self.value = value
        self.color = color

    def __len__(self):
        return len(self.value)

    def _before(self):
        return f"\033[{self.color.value}m"

    def _after(self):
        return "\033[0m"

    def __str__(self):
        return f"{self.color.name}<{self.value}>"

    def __repr__(self):
        return f"{self._before()}{self.value}{self._after()}"

    def __format__(self, format_spec):
        return f"{self._before()}{format(self.value, format_spec)}{self._after()}"


class DiffEnum(Enum):
    ADD = "add"
    REMOVE = "remove"
    CHANGE = "change"


class TableCellDiff:
    def __init__(self,
                 added_value: Optional[FlexTableValue] = None,
                 removed_value: Optional[FlexTableValue] = None):
        self.added_value = added_value
        self.removed_value = removed_value

    def get_diff_type(self):
        if self.added_value and self.removed_value:
            return DiffEnum.CHANGE
        if self.added_value:
            return DiffEnum.ADD
        if self.removed_value:
            return DiffEnum.REMOVE
        if self.added_value is not None:
            return DiffEnum.ADD
        if self.removed_value is not None:
            return DiffEnum.REMOVE
        return None

    def __str__(self):
        return str(self.get_colorized_diff())

    def __repr__(self):
        return repr(self.get_colorized_diff())

    def get_colorized_diff(self):
        diff_type = self.get_diff_type()
        if diff_type == DiffEnum.ADD:
            s = f"+{self.added_value}" if self.added_value != '' else '+""'
            return ColoredString(s, ConsoleColor.GREEN)
        if diff_type == DiffEnum.REMOVE:
            s = f"-{self.removed_value}" if self.removed_value != '' else '-""'
            return ColoredString(s, ConsoleColor.MAGENTA)
        return ColoredString(f"{self.removed_value} -> {self.added_value}", ConsoleColor.YELLOW)

    def __format__(self, format_spec):
        return format(self.get_colorized_diff(), format_spec)




from typing import Optional, List, Union, TextIO
from ..abc import AbstractTestCasesReader
from ..exceptions import FileParsingException
from ..common import zip_columns_with_values
from tabbyset.utils.folder import PathParam
from tabbyset.db.id_utils import is_valid_id
from tabbyset.file_formats.common.parsing_logger import FileParsingLogger
from tabbyset.entities.test_case import TestCase
from tabbyset.utils.flex_table import FlexTable
from tabbyset.file_formats.constants import TEST_CASE_END_LABEL, TEST_CASE_START_LABEL


class Csv1Reader(AbstractTestCasesReader):
    """
    A reader for the CSV1 format.

    Csv1Reader is iterable. All Python iteration methods are supported.
    In order  to iterate over all test cases, you also can use `for` loop:

    >>> reader = Csv1Reader('path/to/file.csv')
    ... for test_case in reader:
    ...     print(test_case)

    :param file: The path of the file or an open file object.
    :param tolerant_mode: If True, the reader will try to read as much as possible from the file.
    :param parsing_logger: A logger for parsing errors.
    """

    def __init__(self, file: Union[PathParam, TextIO],
                 *,
                 tolerant_mode: bool = False,
                 parsing_logger: Optional[FileParsingLogger] = None):
        AbstractTestCasesReader.__init__(self, file, tolerant_mode=tolerant_mode, parsing_logger=parsing_logger)

    def _parse_as_text(self):

        first_column_index = 0

        csvreader = self._prepare_csv_reader()

        current_tc_started = False
        current_tc_ended = False
        current_tc_columns: Optional[List[str]] = None
        current_tc_name: Optional[str] = None
        current_tc_description: Optional[str] = None
        current_tc_id: Optional[str] = None
        current_tc_steps: List[dict[str, str]] = []

        line_number = 0
        test_case_index = -1

        def create_exception(message: str) -> FileParsingException:
            return self._create_reader_exception(message, line_number)

        def log_context() -> dict:
            return {
                'filepath': self._file_path or 'virtual file',
                'lineno': line_number,
                'test_case_index': test_case_index,
                'original_line': str(orig_row)
            }

        def reset_and_get_test_case():
            nonlocal current_tc_name, current_tc_description, current_tc_columns, current_tc_id, current_tc_started, current_tc_ended, current_tc_steps
            if self._parsing_logger:
                if current_tc_name is None:
                    self._parsing_logger.error('Test case name not found', **log_context())
                if current_tc_id is None:
                    self._parsing_logger.info('UUID not found in the test case', **log_context())
            test_case = self._postprocess_test_case(
                test_case=TestCase(name=current_tc_name or 'UNKNOWN',
                                   steps=FlexTable(current_tc_steps),
                                   description=current_tc_description,
                                   id=current_tc_id)
            )

            current_tc_name = None
            current_tc_description = None
            current_tc_columns = None
            current_tc_id = None
            current_tc_started = False
            current_tc_ended = True
            current_tc_steps = []

            return test_case

        for orig_row in csvreader:
            line_number += 1

            row = self._strip_row_right(orig_row)
            row_length = len(row)

            is_name_line = current_tc_name is None and current_tc_started
            is_id_line = current_tc_id is None and current_tc_name is not None
            is_description_line = current_tc_description is None and current_tc_id is not None

            # Skip empty rows
            if ((row_length == 0 or all('' == s for s in row))
                    and not (is_name_line or is_id_line or is_description_line)):
                if self._parsing_logger:
                    if current_tc_columns:
                        self._parsing_logger.info('Empty step row', **log_context())
                continue

            first_row_item = row[first_column_index] if row_length > first_column_index else ''

            if first_row_item == TEST_CASE_START_LABEL:
                test_case_index += 1
                if self._parsing_logger:
                    if row_length != 1:
                        self._parsing_logger.info('Row with the start label has more than 1 item', **log_context())
                if current_tc_started:
                    if self._parsing_logger:
                        self._parsing_logger.error('Started test case is started again', **log_context())
                    if not self._tolerant_mode:
                        raise create_exception('Started test case is started again')
                    else:
                        yield reset_and_get_test_case()
                current_tc_ended = False
                current_tc_started = True
                continue

            if first_row_item not in [TEST_CASE_START_LABEL, TEST_CASE_END_LABEL] and current_tc_started:
                # Lookup for name (always goes right after start label)
                if current_tc_name is None:
                    if self._parsing_logger:
                        if row_length != 1:
                            self._parsing_logger.info('Row with the case name has more than 1 item', **log_context())
                    current_tc_name = first_row_item
                    if not current_tc_name:
                        # This is logged during postprocessing
                        if not self._tolerant_mode:
                            raise create_exception('Test case name not found')
                    continue

                # Lookup for id (always goes right after name)
                # Can be empty
                if is_id_line:
                    # Symbol slot logic is logged during columns reading
                    current_tc_id = first_row_item or ''
                    if self._parsing_logger:
                        is_valid = is_valid_id(current_tc_id)
                        if not current_tc_id:
                            self._parsing_logger.info('UUID is not specified', **log_context())
                        elif not is_valid:
                            self._parsing_logger.info('UUID is not valid', **log_context())

                        if not is_valid:
                            self._parsing_logger.info('UUID will be generated from steps', **log_context())
                    continue

                # Lookup for description (always goes right after id)
                # In most cases empty
                if is_description_line:
                    current_tc_description = first_row_item or ''
                    if self._parsing_logger:
                        if not current_tc_description:
                            self._parsing_logger.debug('Description is not specified', **log_context())
                    continue

                if current_tc_columns is None:
                    current_tc_columns = row
                    if 'Symbol' not in current_tc_columns:
                        if self._parsing_logger:
                            self._parsing_logger.debug('Label "Symbol" not found in the columns', **log_context())
                    continue
                if self._parsing_logger:
                    if row_length > len(current_tc_columns):
                        self._parsing_logger.info('Step row length has more items than columns', **log_context())
                current_tc_steps.append(
                    zip_columns_with_values(current_tc_columns, row)
                )
                continue

            # Return new testcase and reset data
            if first_row_item == TEST_CASE_END_LABEL:
                if not current_tc_started:
                    if self._parsing_logger:
                        self._parsing_logger.error('Not started case tries to end', **log_context())
                    if not self._tolerant_mode:
                        raise create_exception('Not started case tries to end')

                yield reset_and_get_test_case()
        if current_tc_started and not current_tc_ended:
            if self._parsing_logger:
                self._parsing_logger.error('Last test case is not closed', **log_context())
            if not self._tolerant_mode:
                raise create_exception('Last test case is not closed')



from __future__ import annotations

import copy
from typing import Dict, List
from types import MappingProxyType

from .test_case import TestCase


class TestScript:
    name: str
    _test_cases_data: List[TestCase]
    _test_cases_index: Dict[str, TestCase]

    def __init__(self, name: str):
        self.name = name
        self._test_cases_data = []
        self._test_cases_index = {}

    def __copy__(self) -> TestScript:
        new_test_script = TestScript(self.name)
        for tc in self._test_cases_data:
            new_test_script.add_test_case(test_case=tc.copy())
        return new_test_script

    @property
    def test_cases(self) -> MappingProxyType[str, TestCase]:
        return MappingProxyType(self._test_cases_index)

    @property
    def all_columns(self) -> List[str]:
        columns = []
        for tc in self._test_cases_data:
            current_columns = tc.steps.columns
            for column in current_columns:
                if column not in columns:
                    columns.append(column)
        return columns

    def remove_test_case(self, name: str):
        self._test_cases_data.remove(self._test_cases_index[name])
        del self._test_cases_index[name]

    def add_test_case(self, test_case: TestCase):
        self._test_cases_data.append(test_case)
        self._test_cases_index[test_case.name] = test_case

    def copy(self) -> TestScript:
        return copy.copy(self)



from typing import Union
from collections.abc import Iterable
from tabbyset.entities.test_case import TestCase
from tabbyset.utils.flex_table import FlexTable
from tabbyset.file_formats.common.multiheader_csv.config import Categorizer
from tabbyset.file_formats.common.multiheader_csv.core import MultiheaderCsvCore


def _get_global_columns_multiheader(tables: Iterable[Iterable[dict]],
                                    categorizer: Categorizer) -> dict[str, list[str]]:
    categorized_columns: dict[str, dict[str, None]] = {}
    for table in tables:
        if not isinstance(table, FlexTable):
            table = FlexTable(table)
        for row in table:
            category_name = categorizer(row)
            if category_name not in categorized_columns:
                categorized_columns[category_name] = {}
            for column in row:
                if column in categorized_columns[category_name]:
                    continue
                categorized_columns[category_name][column] = None
    return {category_name: list(columns.keys()) for category_name, columns in categorized_columns.items()}


def _get_global_columns_plain(tables: Iterable[Iterable[dict]]) -> list[str]:
    CATEGORY_LABEL = 'UNDEFINED'
    categorized_columns = _get_global_columns_multiheader(tables=tables, categorizer=lambda _: CATEGORY_LABEL)
    return categorized_columns.get(CATEGORY_LABEL, [])


def global_columns(data: Iterable[Union[Iterable[dict], TestCase]],
                   *_,
                   multiheader: bool = False,
                   categorizer: Categorizer = None) -> Union[list[str], dict[str, list[str]]]:
    """
    Extracts global columns from the sequence of test cases or tables.
    :param data: The sequence of test cases or tables.
    :param multiheader: The flag to extract columns with multiheader. Default is False.
    :param categorizer: The function to categorize columns. Default is based on the message type.
    :return: The list of global columns (multiheader=False) or the dictionary of global columns by category (multiheader=True).
    """
    if categorizer is None:
        categorizer = MultiheaderCsvCore.config.categorizer
    tables = (table.steps if isinstance(table, TestCase) else table for table in data)
    if multiheader:
        return _get_global_columns_multiheader(tables, categorizer)
    return _get_global_columns_plain(tables)



"""
Module containing different unit testing utils.
"""
from .test_case import TestCaseAssertions
from .flex_table import FlexTableAssertions
from .ex_generator_testing import ExGeneratorTesting
from .exceptions import TabbySetDiffFail



from .date_range import DateRange
from .folder import Folder, PathParam
from .flex_table import FlexTable, ParsableQueryStatement, DictQuery, sort_with_priority
from .tick_utils import floor_to_tick, ceil_to_tick, round_to_tick, is_multiple_of_tick
from .multi_test_case_writer import MultiTestCaseWriter
from .global_columns import global_columns
from .test_cases_plain_reader import TestCasesPlainReader
from .dhash import dhash
from .fs_utils import walk_tests_folder
from .chunks import chunkify_csv1_file
from .group_by import group_by
from .shuffle import shuffle_csv1, shuffle_csv2


from typing import Optional

from tabbyset.testing.flex_table import FlexTableAssertions
from tabbyset.utils.flex_table.constants import EMPTY_VALUE
from tabbyset.entities import TestCase


def drop_empty_columns(table):
    for column in table.columns:
        if all(value == EMPTY_VALUE for value in table[column]):
            table.remove_column(column)


class TestCaseAssertions(FlexTableAssertions):
    """
    Class containing assertions for the TestCase class.

    Usage:

        >>> class TestMyInstrument(TestCaseAssertions):
        >>>     def test_my_instrument(self):
        >>>         test_case1 = TestCase("My Test Case", [])
        >>>         test_case2 = TestCase("My Test Case", [])
        >>>         self.assertTestCasesEqual(test_case1, test_case2)
    """

    def assertTestCasesEqual(self, tc1: TestCase, tc2: TestCase, msg: Optional[str] = None):
        """
        Assert that two TestCases are equal. In the case of inequality of steps, custom difference is printed.

        :param tc1: The first TestCase.
        :param tc2: The second TestCase.
        :param msg: The message to display on failure.
        """
        self.assertEqual(tc1.name, tc2.name, msg)
        drop_empty_columns(tc1.steps)
        drop_empty_columns(tc2.steps)
        self.assertFlexTablesEqual(
            tc1.steps, tc2.steps,
            msg=f'Actual steps are not equal to the expected ones in the test case "{tc1.name}"! {msg or ""}'
        )



from decimal import Decimal
from typing import Union
from .warnings import get_warn_once_func

Numberish = Union[int, float, Decimal, str]

warn_once = get_warn_once_func()


def _to_decimal(number: Numberish) -> Decimal:
    if isinstance(number, Decimal):
        return number
    if isinstance(number, int):
        return Decimal(number)
    if isinstance(number, float):
        warn_once(f"Converting float ({number}) to Decimal. This may result in loss of precision.")
        return Decimal(str(number))
    return Decimal(str(number))


def is_multiple_of_tick(price: Numberish, tick_size: Numberish) -> bool:
    """
    Checks if the price is a multiple of the tick size.

    All the numbers are converted to Decimal before the operation is performed.

    :param price: The price to check.
    :param tick_size: The tick size to check against.
    :return: True if the price is a multiple of the tick size, False otherwise.
    """
    price = _to_decimal(price)
    tick_size = _to_decimal(tick_size)
    # tick_size == 0 means infinite precision, where all numbers are multiples of each other
    if tick_size == 0:
        return True
    return price % tick_size == 0


def floor_to_tick(price: Numberish, tick_size: Numberish) -> Decimal:
    """
    Rounds the price down to the nearest possible price that is a multiple of the tick size.

    All the numbers are converted to Decimal before the operation is performed.

    :param price: The price to round down.
    :param tick_size: The tick size to round to.
    :return: The rounded down price.
    """
    price = _to_decimal(price)
    tick_size = _to_decimal(tick_size)
    if is_multiple_of_tick(price, tick_size):
        return price
    return price // tick_size * tick_size


def ceil_to_tick(price: Numberish, tick_size: Numberish) -> Decimal:
    """
    Rounds the price up to the nearest possible price that is a multiple of the tick size.

    All the numbers are converted to Decimal before the operation is performed.

    :param price: The price to round up.
    :param tick_size: The tick size to round to.
    :return: The rounded up price.
    """
    price = _to_decimal(price)
    tick_size = _to_decimal(tick_size)
    if is_multiple_of_tick(price, tick_size):
        return price
    return price // tick_size * tick_size + tick_size


def round_to_tick(price: Numberish, tick_size: Numberish) -> Decimal:
    """
    Rounds the price to the nearest possible price that is a multiple of the tick size.

    All the numbers are converted to Decimal before the operation is performed.

    :param price: The price to round.
    :param tick_size: The tick size to round to.
    :return: The rounded price.
    """
    price = _to_decimal(price)
    tick_size = _to_decimal(tick_size)
    if is_multiple_of_tick(price, tick_size):
        return price
    return round(price / tick_size) * tick_size



from uuid import uuid4, uuid5, UUID
from tabbyset.entities.test_case import TestCase

NAMESPACE_TEST_CASE = UUID('c5b52f89-d85e-48cc-b749-dd911b1e7526')

def new_id() -> str:
    """
    Generates a new ID.
    """
    return str(uuid4())

def get_id_from_steps(test_case: TestCase) -> str:
    """
    Generates ID based on the steps of the test case.

    Running this function multiple times with the same test case will always return the same ID.
    """
    return str(uuid5(namespace=NAMESPACE_TEST_CASE,
                     name=str(hash(test_case.steps))))

def is_valid_id(id_string: str) -> bool:
    """
    Checks if the given string is a valid UUID.
    """
    if isinstance(id_string, UUID):
        return True
    if not isinstance(id_string, str):
        return False
    try:
        UUID(id_string)
        return True
    except ValueError:
        return False


from typing import Iterable, Iterator, Union

from tabbyset.entities.test_case import TestCase
from tabbyset.file_formats.csv2.reader import Csv2Reader
from tabbyset.file_formats.abc import AbstractTestCasesReader

class TestCasesPlainReader(Iterable[dict]):
    """
    A utility mapping Test Cases stream to plain CSV stream.
    """

    _test_cases_reader: AbstractTestCasesReader
    _current_test_case: TestCase = None
    _current_steps: Iterator[dict] = None

    def __init__(self, test_cases_reader: AbstractTestCasesReader):
        self._test_cases_reader = test_cases_reader

    @property
    def has_headers(self) -> bool:
        return isinstance(self._test_cases_reader, Csv2Reader)

    @property
    def headers(self) -> Union[list[str], dict[str, list[str]]]:
        if not isinstance(self._test_cases_reader, Csv2Reader):
            raise AttributeError(f"{self._test_cases_reader.__class__.__name__} does not have headers.")
        return self._test_cases_reader.global_columns

    def __iter__(self):
        return self

    def __next__(self):
        if self._current_test_case is None:
            test_case = next(self._test_cases_reader, None)
            if test_case is None:
                raise StopIteration
            self._current_test_case = test_case
            self._current_steps = iter(test_case.steps)
        new_row = next(self._current_steps, None)
        if new_row is None:
            self._current_test_case = None
            return next(self)
        return new_row


from .table_queries import parse_dict_query, QueryStatement, apply_query_to_dict, ParsableQueryStatement, DictQuery
from .flex_table import FlexTable
from .utils import dict_row_to_list, ascii_table, sort_with_priority
from .typing import FlexTableValue, FlexTableRow, TabularData



import logging
import csv

class CSVFileHandler(logging.FileHandler):
    def __init__(self, filename: str):
        super().__init__(filename, mode='w', encoding='utf-8')
        self._initialize_log_file()

    def _initialize_log_file(self):
        writer = csv.writer(self.stream, lineterminator='\n')
        writer.writerow(['level', 'filepath', 'lineno', 'test_case_index', 'summary', 'original_line'])

    def emit(self, record: logging.LogRecord):
        writer = csv.writer(self.stream, lineterminator='\n')
        writer.writerow([
            record.levelname,
            record.pathname,
            record.lineno,
            getattr(record, 'test_case_index', None),
            record.getMessage(),
            getattr(record, 'original_line', None),
        ])

class FileParsingLogger:
    def __init__(self, name: str, filename: str,
                 *,
                 level: int = logging.INFO):
        self.native_logger = logging.getLogger(name)
        is_handler_present = False
        for handler in self.native_logger.handlers:
            if isinstance(handler, CSVFileHandler):
                if handler.baseFilename == filename:
                    is_handler_present = True
                    break
        if not is_handler_present:
            self.native_logger.addHandler(CSVFileHandler(filename))
        self.set_level(level)

    def set_level(self, level: int):
        self.native_logger.setLevel(level)

    def make_record(self, level: int, filepath: str, lineno: int, msg: str, test_case_index: int, original_line: str):
        return self.native_logger.makeRecord(
            name=self.native_logger.name,
            level=level,
            fn=filepath,
            lno=lineno,
            msg=msg,
            args=(),
            exc_info=None,
            extra={'test_case_index': test_case_index, 'original_line': original_line}
        )

    def debug(self, msg: str, filepath: str, lineno: int, test_case_index: int, original_line: str):
        if not self.native_logger.isEnabledFor(logging.DEBUG):
            return
        self.native_logger.handle(self.make_record(logging.DEBUG, filepath, lineno, msg, test_case_index, original_line))

    def info(self, msg: str, filepath: str, lineno: int, test_case_index: int, original_line: str):
        if not self.native_logger.isEnabledFor(logging.INFO):
            return
        self.native_logger.handle(self.make_record(logging.INFO, filepath, lineno, msg, test_case_index, original_line))

    def warning(self, msg: str, filepath: str, lineno: int, test_case_index: int, original_line: str):
        if not self.native_logger.isEnabledFor(logging.WARNING):
            return
        self.native_logger.handle(self.make_record(logging.WARNING, filepath, lineno, msg, test_case_index, original_line))

    def error(self, msg: str, filepath: str, lineno: int, test_case_index: int, original_line: str):
        if not self.native_logger.isEnabledFor(logging.ERROR):
            return
        self.native_logger.handle(self.make_record(logging.ERROR, filepath, lineno, msg, test_case_index, original_line))


import os.path
import sys

from tabbyset.__legacy__.file_formats.v1 import Csv1Reader as LegacyCsv1Reader
import logging
import tabbyset as tbs
import shutil
from tqdm import tqdm

def migrate_legacy_csv1_folder(folder_path: tbs.PathParam,
                               report_path: tbs.PathParam = 'migration_solved_problems.csv',
                               *,
                               deep: bool = False,
                               level: int = logging.INFO):
    legacy_problems_logger = tbs.FileParsingLogger(name=f'csv1_parser/migration/folder/{folder_path}',
                                                   filename=report_path,
                                                   level=level)
    if not os.path.isdir(folder_path):
        raise FileNotFoundError(f'Folder {folder_path} does not exist')
    csv1_pattern = tbs.GlobPatterns.csv1_pattern(deep)
    folder = tbs.Folder(folder_path)
    files = list(folder.glob(csv1_pattern))
    for file in tqdm(files, desc='Migrating Test Scripts', file=sys.stdout):
        tmp_file = str(file)+'.tmp'
        with LegacyCsv1Reader(file, parsing_logger=legacy_problems_logger) as reader, tbs.Csv1Writer(tmp_file) as writer:
            for test_case in reader:
                test_case.id = tbs.TestsTracker.get_id_from_steps(test_case)
                writer.write(test_case)
        shutil.move(tmp_file, file)
    print(f'Migration of {len(files)} files completed')


import copy
from dataclasses import dataclass
from tabbyset.presets.multiheader_configs import msgtype_multiheader_config
from .config import MultiheaderConfig
from typing import Union, Optional

from tabbyset.utils.flex_table import dict_row_to_list
from tabbyset.file_formats.common import complete_row


@dataclass(frozen=True)
class ReadLineResult:
    __slots__ = ('is_header', 'category', 'columns', 'is_header', 'error_msg')
    category: str
    columns: list[str]
    is_header: bool
    error_msg: Optional[str]

class MultiheaderCsvCore:
    config: MultiheaderConfig = msgtype_multiheader_config
    headers: dict[str, list[str]]

    def __init__(self, config: MultiheaderConfig = None):
        if config:
            self.config = config
        self.headers = {}

    def set_headers(self, headers: dict[str, list[str]], writable: bool = False):
        headers = copy.deepcopy(headers)
        self.headers = headers
        if writable:
            for category, columns in headers.items():
                category: str
                columns: list[str]
                if self.config.column_before_row_category is not None:
                    if self.config.column_before_row_category in columns:
                        insert_after = columns.index(self.config.column_before_row_category)
                        columns.insert(insert_after + 1, f'{self.config.row_category_prefix}:{category}')
                    else:
                        columns.append(f'{self.config.row_category_prefix}:{category}')

    def reset_headers(self):
        self.headers = {}

    def read_line(self, line: list[str]) -> ReadLineResult:
        line = list(line)
        err_msg = None
        if len(line) < 2:
            tail = ['', '']
        else:
            tail = line[-2:]
        is_multiheader_columns = (
                tail[0] == 'HeaderDefinition'
                and tail[1].startswith(f'HeaderDefinition{self.config.header_category_postfix}:')
        )
        if is_multiheader_columns:
            category = tail[1].split(':', 1)[1]
            if category in self.headers:
                err_msg = f'Duplicated multiheader category - {category}'
            # Remove technical category column from business data
            row_category_column = f'{self.config.row_category_prefix}:{category}'
            for i, cell in enumerate(line):
                if cell == row_category_column:
                    line[i] = ''
                    break
            self.headers[category] = line[:-2]
            return ReadLineResult(category=category, columns=self.headers[category], is_header=True, error_msg=err_msg)
        else:
            current_row_category = self.get_line_multiheader_category(line)
            if current_row_category is None:
                return ReadLineResult(category='', columns=[], is_header=False, error_msg='Multiheader category not found')
            if current_row_category not in self.headers:
                return ReadLineResult(category=current_row_category, columns=[], is_header=False, error_msg=f'Category "{current_row_category}" is not defined in multiheader')
            return ReadLineResult(category=current_row_category, columns=self.headers[current_row_category], is_header=False, error_msg=None)

    def generate_headers_definitions(self) -> list[list[str]]:
        definitions = []
        for category, columns in self.headers.items():
            tail = [
                'HeaderDefinition',
                f'HeaderDefinition{self.config.header_category_postfix}:{category}',
            ]
            definitions.append(columns + tail)
        return definitions

    def generate_writable_line(self, row: dict) -> list[str]:
        category = self.config.categorizer(row)
        category_string = f'{self.config.row_category_prefix}:{category}'
        current_columns = self.headers.get(category, None)
        if current_columns is None:
            raise ValueError(f'The category "{category}" is not defined in the global columns.')
        return dict_row_to_list(row | {category_string: category_string}, current_columns)

    def check_row_category(self, row: dict, category: str) -> tuple[bool, Optional[str]]:
        category_from_config = self.config.categorizer(row)
        if category_from_config != category:
            return False, f'Category mismatch: expected "{category}", got "{category_from_config}"'
        return True, None

    def get_line_multiheader_category(self, line: list[str]) -> Optional[str]:
        for value in line:
            if value.startswith(f'{self.config.row_category_prefix}:'):
                return value.split(':', 1)[1]
        return None



"""
Module for the business logic utilities around the model.
"""
from .utils import (Folder, PathParam, walk_tests_folder,
                    FlexTable, ParsableQueryStatement, DictQuery, sort_with_priority,
                    chunkify_csv1_file, shuffle_csv1, shuffle_csv2,
                    global_columns, queries,
                    dhash, group_by,
                    floor_to_tick, ceil_to_tick, round_to_tick, is_multiple_of_tick,
                    MultiTestCaseWriter, TestCasesPlainReader)
from .entities import TestCase, TestScript
from .file_formats import (Csv1Reader, Csv1Writer, Csv2Reader, Csv2Writer,
                           FileParsingException, VirtualFileParsingException,
                           MultiheaderConfig, MultiheaderCategorizer,
                           set_default_multiheader_config,
                           MHdrCsvReader, MHdrCsvWriter,
                           RawTestCasesWriter, RawTestCasesReader,
                           FileParsingLogger, GlobPatterns)
from .db.tests_tracker import TestsTracker


"""
Predefined configs for multiheader files.
"""
from tabbyset.file_formats.common.multiheader_csv.config import MultiheaderConfig


def msgtype_categorizer(row: dict) -> str:
    category_key = "MessageType" if "MessageType" in row else "#messagetype"
    return row.get(category_key, 'UNDEFINED')


msgtype_multiheader_config = MultiheaderConfig(row_category_prefix="#messagetype",
                                               column_before_row_category="MessageType",
                                               header_category_postfix="MessageTypes",
                                               categorizer=msgtype_categorizer)



from typing import Sequence, Literal
from tabbyset.utils.folder import PathParam, Folder
from tabbyset.utils.global_columns import global_columns
from tabbyset.file_formats.glob_patterns import GlobPatterns
from tabbyset.entities import TestCase
from tabbyset.file_formats.csv1 import Csv1Reader, Csv1Writer
from tabbyset.file_formats.csv2 import Csv2Reader, Csv2Writer
from .id_utils import new_id, is_valid_id, get_id_from_steps, NAMESPACE_TEST_CASE

FileFormat = Literal['csv1', 'csv2']

class TestsTracker:
    """
    A set of functions for all the most used utilities for operations with tests.
    It also handles the IDs of the tests to keep them consistent.
    """
    NAMESPACE_TEST_CASE = NAMESPACE_TEST_CASE

    new_id = new_id
    is_valid_id = is_valid_id
    get_id_from_steps = get_id_from_steps

    @classmethod
    def new_test(cls,
                 name: str,
                 steps: Sequence[dict],
                 description: str = '' ) -> TestCase:
        """
        Creates a new test with a new ID.
        """
        return TestCase(name=name, steps=steps, description=description, id=cls.new_id())

    @classmethod
    def copy_test(cls, test: TestCase) -> TestCase:
        """
        Creates a copy of the test WITHOUT changing its ID.
        """
        return test.copy()

    @classmethod
    def fork_test(cls, test: TestCase) -> TestCase:
        """
        Creates a copy of the test WITH a new ID.
        """
        test_copy = test.copy()
        test_copy.id = cls.new_id()
        return test_copy

    @classmethod
    def read_file(cls, file_path: PathParam, file_format: FileFormat = 'csv1') -> list[TestCase]:
        """
        Reads tests from a file.
        """
        if file_format == 'csv1':
            with Csv1Reader(file_path) as reader:
                return list(reader)
        elif file_format == 'csv2':
            with Csv2Reader(file_path) as reader:
                return list(reader)
        else:
            raise ValueError(f"Unsupported file format: {file_format}")

    @classmethod
    def read_folder(cls, folder_path: PathParam, file_format: FileFormat = 'csv1', deep=False) -> list[TestCase]:
        """
        Reads tests from a folder.
        """
        folder = Folder(folder_path)
        if file_format == 'csv1':
            test_cases = []
            for file in folder.glob(GlobPatterns.csv1_pattern(deep)):
                with Csv1Reader(file) as reader:
                    test_cases.extend(reader)
            return test_cases
        if file_format == 'csv2':
            test_cases = []
            for file in folder.glob(GlobPatterns.csv2_pattern(deep)):
                with Csv2Reader(file) as reader:
                    test_cases.extend(reader)
            return test_cases
        raise ValueError(f"Unsupported file format: {file_format}")

    @classmethod
    def write_to_file(cls, file_path: PathParam, test_cases: Sequence[TestCase], file_format: FileFormat = 'csv1') -> None:
        """
        Writes tests to a file.
        """
        if file_format == 'csv1':
            with Csv1Writer(file_path) as writer:
                writer.write_many(test_cases)
        elif file_format == 'csv2':
            with Csv2Writer(file_path, global_columns=global_columns(test_cases)) as writer:
                writer.write_many(test_cases)
        else:
            raise ValueError(f"Unsupported file format: {file_format}")


from typing import Union, TextIO, Callable
from .folder import PathParam

from tabbyset.file_formats.csv1.reader import Csv1Reader
from tabbyset.file_formats.csv1.writer import Csv1Writer
from tabbyset.file_formats.csv2.reader import Csv2Reader
from tabbyset.file_formats.csv2.writer import Csv2Writer
from random import Random

FileParam = Union[PathParam, TextIO]

def shuffle_csv1(input_file: FileParam, output_file: FileParam, *, random_seed: int = None) -> None:
    """
    Shuffle the rows of a CSV1 file and write the result to another file.

    By default, the order depends on the random module's random function and by its seed.

    :param input_file: The path of the input CSV1 file or an open file object.
    :param output_file: The path of the output CSV1 file or an open file object.
    :param random_seed: Optional seed for the random number generator to ensure reproducibility.
    """

    with Csv1Reader(input_file) as reader:
        test_cases = list(reader)
        Random(random_seed).shuffle(test_cases)

    with Csv1Writer(output_file, first_priority_columns=[], last_priority_columns=[]) as writer:
        for test_case in test_cases:
            writer.write(test_case)

def shuffle_csv2(input_file: FileParam, output_file: FileParam, *, random_seed: int = None) -> None:
    """
    Shuffle the rows of a CSV2 file and write the result to another file.

    By default, the order depends on the random module's random function and by its seed.

    :param input_file: The path of the input CSV2 file or an open file object.
    :param output_file: The path of the output CSV2 file or an open file object.
    :param random_seed: Optional seed for the random number generator to ensure reproducibility.
    """

    with Csv2Reader(input_file) as reader:
        headers = reader.global_columns
        test_cases = list(reader)
        Random(random_seed).shuffle(test_cases)

    with Csv2Writer(output_file, global_columns=headers) as writer:
        for test_case in test_cases:
            writer.write(test_case)


import copy
import csv
from typing import Optional, Iterable

from ..common.reader import zip_columns_with_values
from ..common.multiheader_csv.config import MultiheaderConfig
from ..common.multiheader_csv.core import MultiheaderCsvCore

class MHdrCsvReader(Iterable[dict[str, str]]):
    """
    A class for reading custom multiheader CSV files.
    """

    _is_headers_read: bool = False
    _multiheader_core: MultiheaderCsvCore
    _unyieled_line: Optional[dict] = None

    def __init__(self, f, dialect="excel",
                 multiheader_config: Optional[MultiheaderConfig] = None, *args, **kwds):
        self.reader = csv.reader(f, dialect, *args, **kwds)
        self._multiheader_core = MultiheaderCsvCore(multiheader_config)
        self.dialect = dialect

    def __iter__(self):
        return self

    @property
    def line_num(self):
        return self.reader.line_num

    @property
    def headers(self):
        if not self._is_headers_read:
            self._unyieled_line = next(self)
        h = copy.deepcopy(self._multiheader_core.headers)
        for k in h:
            h[k] = [x for x in h[k] if x]
        return h


    @headers.setter
    def headers(self, value: dict[str, list[str]]):
        self._multiheader_core.set_headers(value)

    def __next__(self):
        if self._unyieled_line:
            d = self._unyieled_line
            self._unyieled_line = None
            return d
        line, parsed_line = self._parse_next_line()
        if parsed_line.error_msg:
            raise ValueError(parsed_line.error_msg)
        if not self._is_headers_read:
            while parsed_line.is_header:
                line, parsed_line = self._parse_next_line()
            self._is_headers_read = True
        d = zip_columns_with_values(parsed_line.columns, line)
        d.pop('', None)
        return d

    def _parse_next_line(self):
        line = next(self.reader)
        while not line:
            line = next(self.reader)
        return line, self._multiheader_core.read_line(line)




def split_row(row: list[str], columns_length: int):
    return row[:columns_length], row[columns_length:]


def complete_row(row: list[str], columns_length: int):
    return row + [''] * (columns_length - len(row))


def zip_columns_with_values(columns: list[str], values: list[str]) -> dict[str, str]:
    if len(columns) > len(values):
        values = complete_row(values, len(columns))
    return dict(zip(columns, values))



import json
from typing import Union, TextIO
from .tc_to_dict import dict_to_tc
from ..abc import AbstractTestCasesReader
from ..exceptions import FileParsingException
from tabbyset.utils.folder import PathParam


class RawTestCasesReader(AbstractTestCasesReader):
    """
    A reader for the raw test cases stored in JSONL.

    Raw test cases are not limited to specifics of CSV1 or CSV2, but they take much more space.

    RawTestCasesReader is iterable. All Python iteration methods are supported.
    In order  to iterate over all test cases, you also can use `for` loop:

    >>> reader = RawTestCasesReader('path/to/file.csv')
    ... for test_case in reader:
    ...     print(test_case)

    :param file: The path of the file or an open file object.
    """

    def __init__(self, file: Union[PathParam, TextIO]):
        super().__init__(file)

    def _parse_as_text(self):

        jsonl_reader = self._prepare_textio_readable()

        line_number = 0

        def create_exception(message: str) -> FileParsingException:
            return self._create_reader_exception(message, line_number)

        for line in jsonl_reader:
            line_number += 1
            try:
                tc_dict = json.loads(line)
            except json.JSONDecodeError as e:
                raise create_exception(f"Failed to parse JSON: {e}")
            try:
                tc = dict_to_tc(tc_dict)
            except KeyError as e:
                raise create_exception(f"Failed to convert JSON to TestCase: {e}")
            yield tc




from abc import ABC, abstractmethod
from collections.abc import Iterable
from contextlib import AbstractContextManager
from typing import Union, TextIO

from .source_io import SourceIO
from tabbyset.entities.test_case import TestCase
from tabbyset.utils.flex_table import FlexTableRow
from tabbyset.utils.flex_table.utils import dict_row_to_list
from tabbyset.utils.folder import PathParam
from tabbyset.db.id_utils import is_valid_id, new_id


class ITestCasesWriter(AbstractContextManager, ABC):
    @abstractmethod
    def write(self, test_case: TestCase):
        """
        Write a test case to the file.
        :param test_case: The test case to write.
        """
        pass

    @abstractmethod
    def write_many(self, test_cases: Iterable[TestCase]):
        """
        Write sequence of test cases to the file.
        :param test_cases: The list of test cases to write.
        """
        pass

    @abstractmethod
    def close(self):
        """
        Close the file.
        """
        pass


class AbstractTestCasesWriter(SourceIO, ITestCasesWriter, ABC):
    """
    An abstract class for writing test scripts to files.

    :param file: The path of the file or an open file object.
    """

    def __init__(self, file: Union[PathParam, TextIO]):
        SourceIO.__init__(self, file)

    def write(self, test_case: TestCase):
        test_case_to_write = test_case
        if not is_valid_id(test_case.id):
            test_case_to_write = test_case.copy()
            test_case_to_write.id = new_id()
        self._write_test_case(test_case_to_write)

    def write_many(self, test_cases: Iterable[TestCase]):
        for test_case in test_cases:
            self.write(test_case)

    @abstractmethod
    def _write_test_case(self, test_case: TestCase):
        pass  # pragma: no cover

    @staticmethod
    def _table_item_as_list(item: FlexTableRow, columns: list[str]):
        return dict_row_to_list(item, columns)



from tabbyset.__legacy__.file_formats.v1 import Csv1Reader as LegacyCsv1Reader
import tabbyset as tbs
import logging

def legacy_csv1_file_report(file_path: tbs.PathParam,
                            report_path: tbs.PathParam,
                            *,
                            level: int = logging.WARNING) -> None:
    """
    Print the parsing logs for CSV1 file.
    :param file_path: The path to the file.
    :param report_path: The path to the report file.
    :param level: Minimum level of the messages in report. Default: `logging.WARNING`.
    """
    logger = tbs.FileParsingLogger(f'csv1_parser/legacy/file/{str(file_path)}', report_path, level=level)
    with LegacyCsv1Reader(file_path, tolerant_mode=True, parsing_logger=logger) as reader:
        reader.check_validity()

def legacy_csv1_folder_report(folder_path: tbs.PathParam,
                              report_path: tbs.PathParam,
                              *,
                              level: int = logging.WARNING,
                              deep = False):
    csv1_pattern = tbs.GlobPatterns.csv1_pattern(deep)
    folder = tbs.Folder(folder_path)
    logger = tbs.FileParsingLogger(f'csv1_parser/legacy/folder/{str(folder_path)}', report_path, level=level)
    for file in folder.glob(csv1_pattern):
        with LegacyCsv1Reader(file, tolerant_mode=True, parsing_logger=logger) as reader:
            reader.check_validity()



import json
from typing import Union, TextIO

from ..abc.source_io import SourceIO
from tabbyset.entities.test_case import TestCase
from tabbyset.utils.folder import PathParam
from tabbyset.db.id_utils import is_valid_id, get_id_from_steps


class _TCJsonlReader(SourceIO):
    def read_test_case(self) -> TestCase:
        textio = self._prepare_textio_readable()
        testcase = TestCase(name='undefined', steps=[])
        for line in textio:
            raw_value = json.loads(line)
            line_type = raw_value.get('$type', {})
            line_data = raw_value.get('$data', {})
            if line_type == 'meta':
                testcase.name = line_data.get('name', 'undefined')
                testcase.description = line_data.get('description', None)
                testcase.id = line_data.get('id', None)
            elif line_type == 'step':
                testcase.steps.append(line_data)
        self.close()
        if not is_valid_id(testcase.id):
            testcase.id = get_id_from_steps(testcase)
        return testcase

def read_jsonl_testcase(file: Union[PathParam, TextIO]) -> TestCase:
    """
    Read a single test case from a JSONL file.
    """
    with _TCJsonlReader(file) as reader:
        return reader.read_test_case()


from typing import Sequence, Iterable, Union

from tabbyset.utils.flex_table import FlexTableRow


def group_by(table: Sequence[FlexTableRow], by: Union[str, Iterable[str]], *, count: bool=False) -> dict[tuple, Sequence[FlexTableRow]]:
    """
    Groups the rows of a table by the specified columns.

    Args:
        table (Sequence): Table to group.
        by (Iterable): Iterable column names to group by.

    Returns:
        dict[tuple, Sequence]: Mapping from group key (tuple of column values) to FlexTable of rows in that group.
    """
    if not isinstance(table, Sequence):
        raise TypeError("Table must be a sequential object.")
    if not all(isinstance(row, dict) for row in table):
        raise TypeError("Rows in table must be a dictionaries.")
    if not isinstance(by, Iterable):
        raise TypeError("Columns must be an iterable object of column names.")
    if isinstance(by, str):
        by = [by]

    grouped_table = {}

    for idx, row in enumerate(table):
        key = tuple(row.get(col) for col in by)
        if key not in grouped_table:
            grouped_table[key] = 0 if count else []
        grouped_table[key] += 1 if count else [row]
    return grouped_table



import copy
from collections.abc import Iterable, Mapping, Sequence, Callable
from typing import Union, overload, Any

from tabbyset.utils.flex_table.table_queries import parse_dict_query, QueryStatement, apply_query_to_dict, DictQuery
from tabbyset.utils.dhash import dhash
from .typing import FlexTableValue, FlexTableRow
from .utils import flex_table_to_tabular_data, ascii_table
from .constants import EMPTY_VALUE

Query = dict[str, Union[QueryStatement, FlexTableValue]]
Entry = Mapping[str, FlexTableValue]


class FlexTable(Sequence[FlexTableRow]):
    """
    Class representing a table with flexible columns.

    Flexible columns mean that each row can have different columns.

    :param rows: The rows of the table.
    """
    _data: list[FlexTableRow]

    @property
    def columns(self):
        """
        :return: The list of table columns names.
        """
        columns: dict[str, None] = {}
        for row in self._data:
            for key in row:
                if key not in columns:
                    columns[key] = None
        return list(columns.keys())

    @property
    def EMPTY_VALUE(self):
        """
        :return: The value representing an empty cell.
        """
        return EMPTY_VALUE

    @property
    def rows(self):
        """
        :return: The list of table rows.
        """
        return self._data

    def __init__(self, rows: Iterable[FlexTableRow] = None) -> None:
        if rows is None:
            rows = []
        self._data = []
        self.extend(rows)

    def query(self, query: Query) -> 'FlexTable':
        """
        Queries the table with the given query.

        Custom queries are supported and can be used both for `for` loops and `in` statements.

            >>> table: FlexTable

        - If <query> in FlexTable:
            >>> if {"Action": "Quote"} in table:
            >>>     # Do something, if there is a row with "Action" equal to "Quote"
        - For row in FlexTable[<query>]:
            >>> for row in table.query({"Action": "Quote"}):
            >>>     # Do something with each row with "Action" equal to "Quote"

        **Numbers**

        All the string values are converted to numbers if possible. So the following queries are equal:
            >>> table.query({"Price": "100"})
            >>> table.query({"Price": 100})

        Therefore, query statements can be used with numbers as well:
            >>> table.query({"Price": "> 100"})
            >>> table.query({"Price": "> 100.0"})

        As comparison of the float numbers is not always precise, it is recommended to use strings for float numbers:
            >>> table.query({"Price": "100.0"})

        **Query statements**

        Query statements can be:

        - Equal value (`"<value>"`, `"= <value>"`)
            >>> table.query({"Action": "Quote"})
            >>> table.query({"Action": "= Quote"})
        - Not equal value (`"!= <value>"`)
            >>> table.query({"Action": "!= Quote"})
        - Greater than value (`"> <value>"`)
            >>> table.query({"Price": "> 100"})
        - Greater than or equal value (`">= <value>"`)
            >>> table.query({"Price": ">= 100"})
        - Less than value (`"< <value>"`)
            >>> table.query({"Price": "< 100"})
        - Less than or equal value (`"<= <value>"`)
            >>> table.query({"Price": "<= 100"})

        **Multiple keys**

        If query dictionary contains multiple keys, all of them should be satisfied (acts as `AND` statement).

        For example, the following will return all rows with "Action" equal to "Quote" and "Price" greater than 100.:
            >>> table.query({"Action": "Quote", "Price": "> 100"})
        """
        return self._query(query)

    def _query(self, query: DictQuery) -> 'FlexTable':
        if isinstance(query, dict):
            query = parse_dict_query(query)
            return FlexTable([row for row in self._data if apply_query_to_dict(query, row)])
        raise TypeError(f"Invalid query type: {type(query)}")

    def column_values(self, column: str) -> list[FlexTableValue]:
        """
        :return: The list of values in the given column.
        """
        return self._column_values(column)

    def _column_values(self, column: str) -> list[FlexTableValue]:
        if not isinstance(column, str):
            raise TypeError(f"Invalid column name type: {type(column)}")
        return [row.get(column, EMPTY_VALUE) for row in self._data]

    def contains(self, query: DictQuery) -> bool:
        """
        Checks if the table contains rows matching with the given query.
        """
        return self.__contains__(query)

    def insert(self, idx: int, row: Entry) -> 'FlexTable':
        """
        Inserts a row to the table at the given index.
        returns: The table itself.
        """
        self._data.insert(idx, self._format_entry(row))
        return self

    def append(self, row: Entry) -> 'FlexTable':
        """
        Appends a row to the table.
        """
        self._data.append(self._format_entry(row))
        return self

    def extend(self, rows: Iterable[Entry]) -> 'FlexTable':
        """
        Extends the table with the given rows.
        """
        for row in rows:
            self.append(row)
        return self

    def count(self, query: Query) -> int:
        """
        :return: The number of rows matching the given query.
        """
        return len(self.query(query))

    def index(self, query: Query, start = 0, stop: int = None) -> int:
        """
        :return: The index of the first occurrence of the value in the table.

        If the value is not found, a ValueError is raised.
        """
        for i in range(start, stop or len(self)):
            if apply_query_to_dict(parse_dict_query(query), self[i]):
                return i
        raise ValueError(f"Value not found: {query}")

    def pop(self, idx: int = -1) -> FlexTableRow:
        """
        Removes the row at the given index (default last) and returns it.
        """
        return self._data.pop(idx)

    def remove(self, query: Query) -> 'FlexTable':
        """
        Removes the rows matching the given query.
        """
        self._data = [row for row in self._data if not apply_query_to_dict(parse_dict_query(query), row)]
        return self

    def clear(self) -> None:
        """
        Clears the table.
        """
        self._data.clear()

    def reverse(self) -> 'FlexTable':
        """
        Reverses the table.
        """
        self._data.reverse()
        return self

    def sort(self, *, key: Callable[[FlexTableRow], Any]=None, reverse=False) -> 'FlexTable':
        """
        Sorts the table.
        """
        self._data.sort(key=key, reverse=reverse)
        return self

    def remove_column(self, column: str) -> None:
        """
        Removes the given column from all the rows of the table.
        """
        self.remove_columns((column,))

    def remove_columns(self, columns: Iterable[str]) -> None:
        """
        Removes the given columns from all the rows of the table.
        """
        for row in self._data:
            for column in columns:
                row.pop(column, None)

    def copy(self, deep: bool = False) -> 'FlexTable':
        """
        :return: A copy of the table.
        """
        if deep:
            return copy.deepcopy(self)
        return copy.copy(self)

    def as_ascii_table(self):
        """
        :return: The table as an ASCII table to be printed in the console.
        """
        return ascii_table(flex_table_to_tabular_data(self.columns, self))

    def __eq__(self, other):
        if not isinstance(other, FlexTable):
            return NotImplemented
        return sorted(self.columns) == sorted(other.columns) and tuple(self) == tuple(other)

    def __repr__(self):
        return f'FlexTable({self._data})'

    def __len__(self):
        return len(self._data)

    def __setitem__(self, idx: Union[int, str], value):
        if isinstance(idx, int):
            self._data[idx] = dict(value)
            return
        if isinstance(idx, str):
            for row in self._data:
                row[idx] = value
            return
        raise TypeError("Invalid index type")

    @overload
    def __getitem__(self, idx: int) -> FlexTableRow:
        pass

    @overload
    def __getitem__(self, idx: slice) -> 'FlexTable':
        pass

    @overload
    def __getitem__(self, idx: str) -> list[FlexTableValue]:
        pass

    @overload
    def __getitem__(self, idx: DictQuery) -> 'FlexTable':
        pass

    def __getitem__(self,
                    idx: Union[int, slice, str, FlexTableRow, DictQuery]
                    ) -> Union[FlexTableRow, 'FlexTable', list[FlexTableValue]]:
        if isinstance(idx, int):
            return self._data[idx]
        if isinstance(idx, slice):
            return FlexTable(self._data[idx])
        if isinstance(idx, str):
            return self._column_values(idx)
        if isinstance(idx, dict):
            return self._query(idx)
        raise TypeError(f"Invalid index type: {type(idx)}")

    def __delitem__(self, idx: Union[int, str]):
        if isinstance(idx, int):
            del self._data[idx]
            return
        if isinstance(idx, str):
            self.remove_column(idx)
            return

    def __iter__(self):
        return iter(self._data)

    def __contains__(self, item: DictQuery) -> bool:
        query = parse_dict_query(item)
        for it in self._data:
            if apply_query_to_dict(query, it):
                return True

    def __copy__(self):
        new_instance = FlexTable(self._data.copy())
        return new_instance

    def __deepcopy__(self, memodict=None):
        new_instance = FlexTable(copy.deepcopy(self._data, memodict))
        return new_instance

    def __add__(self,
                other: Union[
                    'FlexTable',
                    Entry,
                    Iterable[Entry]
                ]) -> 'FlexTable':
        new_instance = self.copy()
        self._add_generic(new_instance, other)
        return new_instance

    def __iadd__(self, other: Union[
        'FlexTable',
        Entry,
        Iterable[Entry]
    ]):
        self._add_generic(self, other)
        return self

    def __hash__(self) -> int:
        columns = self.columns
        meaningful_columns = set(columns)
        for column in set(columns):
            if not any(self.column_values(column)):
                meaningful_columns.remove(column)
        meaningful_columns = sorted(meaningful_columns)
        tabular_data = flex_table_to_tabular_data(meaningful_columns, self)
        tabular_data = tuple(tuple(row) for row in tabular_data)
        return dhash(tabular_data)

    @staticmethod
    def _add_generic(table: 'FlexTable', other: Union[
        'FlexTable',
        Entry,
        Iterable[Entry]
    ]) -> 'FlexTable':
        if isinstance(other, FlexTable):
            return table.extend(other)
        if isinstance(other, Mapping):
            return table.append(other)
        if isinstance(other, Iterable):
            return table.extend(other)
        raise TypeError(f"Invalid type: {type(other)}")

    @staticmethod
    def _format_entry(entry: Entry) -> dict:
        if not isinstance(entry, dict):
            entry = dict(entry)
        return entry



from typing import Type, Optional, TextIO
from warnings import warn_explicit
from functools import wraps
import sys


def _get_caller_filename_and_lineno():
    import inspect
    frame = inspect.currentframe()
    while frame:
        if frame.f_globals.get('__name__') != __name__:
            return frame.f_code.co_filename, frame.f_lineno
        frame = frame.f_back
    return "<unknown>", 0


def libwarn(message, category: Type[Warning] = RuntimeWarning):
    filename, lineno = _get_caller_filename_and_lineno()
    warn_explicit(message=message, category=category, filename=filename, lineno=lineno)


def warn_with_traceback(message: str,
                        category: Type[Warning],
                        filename: str,
                        lineno: int,
                        file: Optional[TextIO] = None):
    import traceback
    log = file if hasattr(file, 'write') else sys.stderr
    log.write(f"{filename}:{lineno}: {category.__name__}: {message}\n")
    traceback.print_stack(file=log)
    log.write("\n")


def get_warn_once_func():
    is_warned = False

    def warn_once(message, category: Type[Warning] = RuntimeWarning):
        nonlocal is_warned
        if not is_warned:
            is_warned = True
            libwarn(message, category)

    return warn_once


def get_warn_once_per_line_func():
    warned_lines = set()

    def warn_once(message, category: Type[Warning] = RuntimeWarning):
        filename, lineno = _get_caller_filename_and_lineno()
        if (filename, lineno) not in warned_lines:
            warned_lines.add((filename, lineno))
            warn_explicit(message=message, category=category, filename=filename, lineno=lineno)

    return warn_once


def experimental_method(func):
    warn_once = get_warn_once_per_line_func()

    @wraps(func)
    def wrapper(*args, **kwargs):
        class_info = args[0].__class__
        warn_once(f"{class_info.__name__}.{func.__name__} is experimental and may change in the future.",
                  category=FutureWarning)
        return func(*args, **kwargs)

    return wrapper



class GlobPatterns:
    """
    Class for defining the glob patterns for the file formats.
    """

    @classmethod
    def _deepify(cls, pattern: str) -> str:
        return '**/' + pattern

    @classmethod
    def csv1_pattern(cls, deep=False) -> str:
        pattern = '*.csv'
        if deep:
            pattern = cls._deepify(pattern)
        return pattern

    @classmethod
    def csv2_pattern(cls, deep=False) -> str:
        pattern = '*.matrix.csv'
        if deep:
            pattern = cls._deepify(pattern)
        return pattern


import json

from ..abc.source_io import SourceIO

class TcJsonlWriter(SourceIO):
    """
    Implementation for writing a single testcase in JSONL format.

    Might be useful for simultaneous writing of several testcases without the need to keep all of them in memory.

    :param file: The path of the file or an open file object.
    """

    def write_metadata(self, metadata: dict):
        """
        Write the metadata of the testcase.

        Obligatory to write before writing any steps.

        Mandatory fields: name
        Optional fields: description

        :param metadata: The metadata to write.
        """
        textio = self._prepare_textio_writable()
        textio.write(json.dumps({ '$type': 'meta', '$data': metadata }))
        textio.write('\n')

    def write_step(self, step: dict):
        """
        Write a step of the testcase.

        :param step: The step to write.
        """
        textio = self._prepare_textio_writable()
        textio.write(json.dumps({ '$type': 'step', '$data': step }))
        textio.write('\n')


from tabbyset.utils.folder import PathParam


class FileParsingException(Exception):
    def __init__(self, file_path: PathParam, line_number: int, message: str):
        """
        Custom exception for parsing errors.

        :param file_path: Path to file trying to parse.
        :param line_number: The line number where the parsing error occurred.
        :param message:  A description of the parsing error.
        """
        super().__init__(message)
        self.line_number = line_number
        self.file_path = file_path

    def __str__(self):
        return f"{super().__str__()}: {self.file_path}:{self.line_number}"


class VirtualFileParsingException(FileParsingException):
    def __init__(self, file: str, line_number: int, message: str):
        Exception.__init__(self, message)
        self.line_number = line_number
        self.file = file

    def __str__(self):
        return f"{Exception.__str__(self)} on line {self.line_number} \n{self.file}"



from typing import Union, TextIO
from ..abc import AbstractTestCasesWriter
from ...utils.flex_table.utils import sort_with_priority
from tabbyset.utils.folder import PathParam
from tabbyset.file_formats.constants import TEST_CASE_END_LABEL, TEST_CASE_START_LABEL
from tabbyset.entities.test_case import TestCase


class Csv1Writer(AbstractTestCasesWriter):
    """
    A writer for CSV1 test scripts.
    :param file: The path of the file or an open file object.
    :param first_priority_columns: The columns that should be written first. Default: `['Status', 'ID', 'PreviousID', 'Action', 'User', 'Symbol', 'Side', 'OrderType', 'TIF', 'OrderQty', 'Price']`.
    :param last_priority_columns: The columns that should be written last. Default: `[]`.
    """

    # Default values are stored in the tuple to prevent modification of the default values.
    _first_priority_columns: list[str]
    _default_first_priority_columns = ('Status', 'ID', 'PreviousID', 'Action', 'User', 'Symbol', 'Side',
                                       'OrderType', 'TIF', 'OrderQty', 'Price')
    _last_priority_columns: list[str]
    _default_last_priority_columns: tuple[str] = tuple()

    def __init__(self,
                 file: Union[PathParam, TextIO],
                 *_,
                 first_priority_columns: list[str] = None,
                 last_priority_columns: list[str] = None):
        AbstractTestCasesWriter.__init__(self, file)
        if first_priority_columns is not None:
            self._first_priority_columns = first_priority_columns
        else:
            self._first_priority_columns = list(self._default_first_priority_columns)
        if last_priority_columns is not None:
            self._last_priority_columns = last_priority_columns
        else:
            self._last_priority_columns = list(self._default_last_priority_columns)

    def _write_test_case(self, test_case: TestCase):
        writer = self._prepare_csv_writer()

        test_case_columns = sort_with_priority(test_case.steps.columns,
                                               first_items=self._first_priority_columns,
                                               last_items=self._last_priority_columns)

        writer.writerow([TEST_CASE_START_LABEL])
        writer.writerow([test_case.name])
        writer.writerow([test_case.id])
        writer.writerow([test_case.description] if test_case.description else [])
        writer.writerow(test_case_columns)
        for step in test_case.steps:
            writer.writerow(self._table_item_as_list(step, test_case_columns))
        writer.writerow([TEST_CASE_END_LABEL])
        writer.writerow([])
        writer.writerow([])



def run_select_subset_sqlite(venue: str,
                             folder_traces,
                             folder_cases,
                             coverage_script: str,
                             folder_output,
                             *,
                             keep_db: bool = False,
                             with_code_coverage: bool = False,
                             iterator_type):
    """
    This function runs the select subset pipeline.

    In the result of this pipeline, following files will be saved in the `folder_output`:

    - `traces.db`: The SQLite database containing the data used.
    - `<coverage_script>.maxanno.csv`: Minimum set of test cases that cover all traces merged into single CSV1 file.
    - `<coverage_script>.minanno.csv`: Maximum set of test cases that cover all traces merged into single CSV1 file.
    - `<venue>_annotations_summary.csv`: Report file containing the summary of all traces.
    - `<venue>_duplicates_summary.csv`: Report file containing links of traces covered by the same set of test cases.
    - `weights.csv`: Report on the chosen test cases (`1` means test case is chosen, `0` otherwise).

    :param venue: The venue name, used in the names of report files.
    :param folder_traces: The folder containing the traces.
    :param folder_cases: The folder containing the test cases.
    :param coverage_script: The name of the coverage script, used in the names of CSV1 files.
    :param folder_output: The folder where the output files will be saved.
    :param keep_db: If True, the SQLite database will not be deleted after the pipeline finishes.
    :param with_code_coverage: If True, additional traces will be created from .coverage.json files.
    :param iterator_type: The type of iterator to use when selecting the subset.
        `IteratorType.COUNT_ITERATOR` sorts test_cases depending on the number of traces they cover.
        `IteratorType.WEIGHTS_ITERATOR` sorts test_cases depending on the sum of the weights of the traces they cover.
        Default is `IteratorType.COUNT_ITERATOR`.
    """
    raise ModuleNotFoundError(
        'Select subset SQLite was moved to the proprietary module since TabbySet went Open Source. '
        'Use "from ex_subset_selector.select_subset_sqlite import run_select_subset_sqlite".')



import copy

from typing import Union, TextIO, Optional
from ..abc import AbstractTestCasesWriter
from tabbyset.file_formats.constants import TEST_CASE_END_LABEL, TEST_CASE_START_LABEL
from tabbyset.entities.test_case import TestCase
from ..common import complete_row
from tabbyset.file_formats.common.multiheader_csv import MultiheaderConfig
from tabbyset.file_formats.common.multiheader_csv.core import MultiheaderCsvCore
from tabbyset.utils.warnings import libwarn
from tabbyset.utils.folder import PathParam


class Csv2Writer(AbstractTestCasesWriter):
    """
    A writer for CSV2 test scripts.

    In CSV2 format, the global columns are written only once at the beginning of the file.
    It is necessary to provide the list of global columns to the writer.

    :param file: The path of the file or an open file object.
    :param global_columns: Superset of columns for all test cases in the file. In case of multiheader, it is a dictionary with category names as keys and lists of columns as values.
    :param multiheader: The flag to specify explicitly if you want use multiheader or not. Default is None, what means that the writer will decide automatically.
    :param multiheader_config: The configuration for multiheader. Default is the message type based config.
    """
    _global_columns = []
    _global_columns_written = False
    _multiheader = False
    _multiheader_core: MultiheaderCsvCore

    def __init__(self,
                 file: Union[PathParam, TextIO],
                 global_columns: Union[list[str], dict[str, list[str]]],
                 *_,
                 multiheader: Optional[bool] = None,
                 multiheader_config: Optional[MultiheaderConfig] = None):
        AbstractTestCasesWriter.__init__(self, file)
        self._global_columns = copy.deepcopy(global_columns)
        auto_multiheader = isinstance(global_columns, dict)
        if multiheader is None:
            multiheader = auto_multiheader
        elif multiheader != auto_multiheader:
            if auto_multiheader:
                raise ValueError('The global columns are a dictionary, so the multiheader flag should be None or True.')
            else:
                raise ValueError('The global columns are a list, so the multiheader flag should be None or False.')
        self._multiheader = multiheader
        self._multiheader_core = MultiheaderCsvCore(multiheader_config)
        if self._multiheader:
            self._multiheader_core.set_headers(self._global_columns, writable=True)
        if multiheader_config:
            if not multiheader:
                libwarn('The multiheader_config parameter is ignored because the multiheader is not used.',
                        category=RuntimeWarning)

    @classmethod
    def get_default_multiheader_config(cls) -> MultiheaderConfig:
        return MultiheaderCsvCore.config

    @classmethod
    def set_default_multiheader_config(cls, multiheader_config: MultiheaderConfig):
        MultiheaderCsvCore.config = multiheader_config

    @property
    def _max_cells_in_row(self):
        if self._multiheader:
            return max(len(columns) + 2 for columns in self._global_columns.values())
        return len(self._global_columns)

    def _get_tc_start_row(self, test_case_name: str):
        return complete_row([TEST_CASE_START_LABEL, test_case_name], self._max_cells_in_row)

    def _get_tc_end_row(self):
        return complete_row([TEST_CASE_END_LABEL], self._max_cells_in_row)

    def _write_test_case(self, test_case: TestCase):
        # TODO: Write test case id
        if self._multiheader:
            self._write_test_case_multiheader(test_case)
        else:
            self._write_test_case_classic(test_case)

    def _write_test_case_classic(self, test_case: TestCase):
        writer = self._prepare_csv_writer()

        if not self._global_columns_written:
            writer.writerow(self._global_columns)
            self._global_columns_written = True

        writer.writerow(self._get_tc_start_row(test_case.name))
        writer.writerow([(common_column if (common_column in test_case.steps.columns) else '')
                         for common_column in self._global_columns])
        for step in test_case.steps:
            writer.writerow(self._table_item_as_list(step, self._global_columns))
        writer.writerow(self._get_tc_end_row())

    def _write_test_case_multiheader(self, test_case: TestCase):
        writer = self._prepare_csv_writer()

        if not isinstance(self._global_columns, dict):
            raise ValueError('The global columns should be a dictionary for multiheader.')

        if not self._global_columns_written:
            writer.writerows(self._multiheader_core.generate_headers_definitions())
            self._global_columns_written = True

        writer.writerow(self._get_tc_start_row(test_case.name))
        for step in test_case.steps:
            writer.writerow(
                complete_row(self._multiheader_core.generate_writable_line(step),
                             self._max_cells_in_row)
            )
        writer.writerow(self._get_tc_end_row())



import unittest
import logging
from typing import Optional
from tabbyset.utils.flex_table import FlexTable, ascii_table
from .exceptions import TabbySetDiffFail
from .flex_table_diff import get_flex_tables_diff, get_columns_diff
from .diff import TableCellDiff, ColoredString

logger = logging.getLogger(__name__)

class FlexTableAssertions(unittest.TestCase):
    """
    Class containing assertions for the FlexTable class.

    Usage:

        >>> class TestMyInstrument(FlexTableAssertions):
        >>>     def test_my_instrument(self):
        >>>         table1 = FlexTable()
        >>>         table2 = FlexTable()
        >>>         self.assertFlexTablesEqual(table1, table2)
    """
    def assertFlexTablesEqual(self, table1: FlexTable, table2: FlexTable, msg: Optional[str] = None):
        """
        Assert that two FlexTables are equal. In the case of inequality, custom difference is printed.

        :param table1: The first FlexTable.
        :param table2: The second FlexTable.
        :param msg: The message to display on failure.
        """
        if table1 == table2:
            return
        try:
            columns_diff, merged_columns = get_columns_diff(table1, table2)
            table_diff = get_flex_tables_diff(table1, table2, merged_columns)

            table_data = [columns_diff] + table_diff
            table_width_ref = [
                [
                    len(cell.get_colorized_diff().value) if isinstance(cell, TableCellDiff)
                    else len(cell.value) if isinstance(cell, ColoredString)
                    else len(cell) for cell in row
                ]
                for row in table_data
            ]
        except Exception as e:
            raise TabbySetDiffFail(e)
        if msg:
            added_msg = f"\n: {msg}"
        else:
            added_msg = ""
        self.fail(f"Tables are not equal:\n{ascii_table(table_data, table_width_ref)}{added_msg}")



from abc import ABC, abstractmethod
from collections.abc import Iterator, Iterable
from typing import Optional, Union, Generator, TextIO

from tabbyset.db.id_utils import get_id_from_steps, is_valid_id
from .source_io import SourceIO
from ..exceptions import FileParsingException, VirtualFileParsingException
from tabbyset.file_formats.common.parsing_logger import FileParsingLogger
from tabbyset.entities.test_case import TestCase
from tabbyset.utils.folder import PathParam


class AbstractTestCasesReader(SourceIO, Iterable[TestCase], ABC):
    """
    An abstract class for reading test scripts from files.

    :param file: The path of the file or an open file object.
    """
    _iterator: Optional[Iterator[TestCase]] = None
    _is_iterator_done: bool = False
    _tolerant_mode: bool

    def __init__(self,
                 file: Union[PathParam, TextIO],
                 *,
                 tolerant_mode: bool = False,
                 parsing_logger: Optional[FileParsingLogger] = None):
        SourceIO.__init__(self, file)
        self._tolerant_mode = tolerant_mode
        self._parsing_logger = parsing_logger

    @abstractmethod
    def _parse_as_text(self) -> Generator[TestCase, None, None]:
        pass  # pragma: no cover

    def read_all(self):
        """
        Read all test cases from the file.
        :return: A list of test cases.
        """
        return list(iter(self))

    def read_one(self) -> Optional[TestCase]:
        """
        Read a next test case from the file.

        Updates the internal iterator.

        :return: A test case.
        """
        return next(self, None)

    def restart_reading(self):
        """
        Restart reading from the beginning of the file.
        """
        current_iterator = self._iterator
        if current_iterator is not None:
            del current_iterator
            self._iterator = None
        if self._textio is not None:
            self._textio.seek(self._starting_position)
        self._iterator = self._parse_as_text()

    def check_validity(self) -> bool:
        """
        Check if the file is valid.

        This method restarts reading from the beginning of the file.
        :return: True if the file is valid, False otherwise.
        """
        try:
            self.restart_reading()
            for _ in iter(self):
                pass
            self.restart_reading()
            return True
        except FileParsingException as e:
            self.restart_reading()
            return False

    def __iter__(self) -> Iterator[TestCase]:
        """
        Iterate over test cases in the file.
        :return: An iterator over test cases.
        :raises tabbyset.FileParsingException: If the file is not valid.
        :raises tabbyset.VirtualFileParsingException: If the files content is not valid and filepath is not available.
        """
        if self._iterator is None:
            self.restart_reading()
        return self._iterator

    def __next__(self) -> TestCase:
        return next(iter(self))

    def _create_reader_exception(self, message: str, line_number: int) -> FileParsingException:
        file = self._file_path
        if file:
            return FileParsingException(file_path=file, line_number=line_number, message=message)
        if self._textio is not None:
            self._textio.seek(0)
            file = self._textio.read()
        return VirtualFileParsingException(file=file, line_number=line_number, message=message)

    @staticmethod
    def _postprocess_test_case(test_case: TestCase) -> TestCase:
        if '' in test_case.steps.columns:
            test_case.steps.remove_column('')
        if not is_valid_id(test_case.id or ''):
            test_case.id = get_id_from_steps(test_case)
        return test_case

    @staticmethod
    def _strip_row_right(row: list[str]) -> list[str]:
        for i in range(len(row) - 1, -1, -1):
            if row[i]:
                return row[:i + 1]
        return []



import os
import shutil
from pathlib import Path
from inspect import currentframe, getframeinfo
from typing import Union

PathParam = Union[os.PathLike, str, Path]


class Folder(os.PathLike):
    """
    A class to represent a folder in the file system.

    :param path: The path of the folder.
    """
    path: Path

    def __init__(self, path: PathParam):
        if not os.path.isabs(path):
            path = (Path(os.path.dirname(os.getcwd() + '/')) / path).resolve()
        if not os.path.exists(path) and not os.path.isdir(path):
            os.makedirs(path)
        self.path = Path(path)

    def listdir(self):
        """
        List all files and folders in the folder.
        :return: A list of file and folder names.
        """
        return os.listdir(self.path)

    def glob(self, pattern: str):
        """
        Iterate over this subtree and yield all existing files (of any kind, including directories) matching the given relative pattern.
        :param pattern: The pattern to match.
        :return: An iterator over the matching file path's.
        """
        return self.path.glob(pattern)

    def get_file_path(self, file_name: str) -> Path:
        """
        Get the path of a file in the folder.
        """
        if os.path.isabs(file_name):
            raise ValueError(f'The file name should be relative to the folder, got an absolute path instead {file_name}')
        return self.path / file_name

    def is_file_exists(self, file_name: str) -> bool:
        """
        Check if a file exists in the folder.
        :param file_name: The name of the file.
        :return: True if the file exists, False otherwise.
        """
        return self.get_file_path(file_name).exists()

    def clear(self):
        """
        Remove all files and folders in the folder.
        """
        shutil.rmtree(self.path)
        os.makedirs(self.path)

    def mount_subfolder(self, subfolder: str) -> 'Folder':
        """
        Mount a subfolder in this folder.
        :param subfolder: The name of the subfolder.
        :return: The subfolder.
        """
        return Folder(self.path / subfolder)

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return str(self.path)

    def __fspath__(self):
        return str(self.path)

    @classmethod
    def mount_from_current_module(cls, folder_rel_path: PathParam) -> 'Folder':
        """
        Mount a folder relative to the current module.

        Might be useful when you want to call the script from different directories.

        :param folder_rel_path: The relative path of the folder.
        :return: The folder.
        """
        current_file_path = getframeinfo(currentframe().f_back).filename
        path = (Path(os.path.dirname(current_file_path)) / folder_rel_path).resolve()
        return cls(path)



import csv

from contextlib import AbstractContextManager
from abc import ABC
from os import PathLike
from pathlib import Path
from io import TextIOBase
from typing import Optional, Union, TextIO
from tabbyset.utils.folder import PathParam


class SourceIO(AbstractContextManager, ABC):
    """
    Implementation for source input output operations.

    :param file: The path of the file or an open file object.
    """
    _file_path: Optional[PathParam] = None
    _textio: Optional[TextIO] = None
    _starting_position: int = 0

    def __init__(self, file: Union[PathParam, TextIO]):
        if isinstance(file, TextIOBase):
            self._textio = file
            self._starting_position = file.tell()
        elif isinstance(file, (PathLike, str, Path)):
            self._file_path = file
        else:
            raise ValueError(f"Invalid file provided: {file}")

    def _prepare_textio_writable(self):
        if self._textio is None:
            self._textio = open(self._file_path, "w", newline='', encoding='utf-8')
        return self._textio

    def _prepare_textio_readable(self):
        if self._textio is None:
            self._textio = open(self._file_path, "r", newline='', encoding='utf-8')
        else:
            self._textio.seek(self._starting_position)
        return self._textio

    def _prepare_csv_writer(self):
        return csv.writer(self._prepare_textio_writable())

    def _prepare_csv_reader(self):
        return csv.reader(self._prepare_textio_readable())

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()

    def close(self):
        """
        Close the file.
        """
        if self._textio is not None:
            self._textio.close()



from .constants import EMPTY_VALUE
from .typing import FlexTableValue, FlexTableRow, TabularData
from collections.abc import Iterable


def dict_row_to_list(row: FlexTableRow, ordered_columns: list[str]) -> list[FlexTableValue]:
    return [row.get(col, EMPTY_VALUE) for col in ordered_columns]


def flex_table_to_tabular_data(columns: list[str], rows: Iterable[FlexTableRow]) -> TabularData:
    tabular_data = [columns]
    for row in rows:
        tabular_data.append(dict_row_to_list(row, columns))
    return tabular_data


def sort_with_priority(items: list[str], *_, first_items: list[str] = None, last_items: list[str] = None) -> list[str]:
    """
    Sorts a list of items with priority.
    The first_items will be placed first, the last_items will be placed last in the provided order.
    :param items: The list of items to sort.
    :param first_items: The items to place first.
    :param last_items: The items to place last.
    """
    if first_items is None:
        first_items = []
    if last_items is None:
        last_items = []

    middle_result_items = list(items)

    first_result_items = []
    for item in first_items:
        if item in items:
            first_result_items.append(item)
            middle_result_items.remove(item)

    last_result_items = []
    for item in last_items:
        if item in items:
            last_result_items.append(item)
            middle_result_items.remove(item)

    return first_result_items + middle_result_items + last_result_items


def ascii_table(data: TabularData, ref_widths_data: list[list[int]] = None) -> str:
    if ref_widths_data is None:
        column_widths = [max(len(str(item)) for item in col) for col in zip(*data)]
    else:
        column_widths = [max(col) for col in zip(*ref_widths_data)]
    result = []
    # Print the data in a tabular format
    for i, row in enumerate(data):
        result.append(" | ".join("{:<{}}".format(item, width) for item, width in zip(row, column_widths)))
        if i == 0:
            result.append("=|=".join("=" * col_w for col_w in column_widths) + "=")
    return "\n".join(result)



import json

from typing import Union, TextIO
from .tc_to_dict import tc_to_dict
from ..abc import AbstractTestCasesWriter
from tabbyset.entities.test_case import TestCase
from tabbyset.utils.folder import PathParam


class RawTestCasesWriter(AbstractTestCasesWriter):
    """
    A writer for test scripts in raw JSONL format.

    It is not limited to specifics of CSV1 or CSV2, but takes much more space.
    :param file: The path of the file or an open file object.
    """

    # Default values are stored in the tuple to prevent modification of the default values.

    def __init__(self, file: Union[PathParam, TextIO]):
        AbstractTestCasesWriter.__init__(self, file)

    def _write_test_case(self, test_case: TestCase):
        writer = self._prepare_textio_writable()

        writer.write(json.dumps(tc_to_dict(test_case)))
        writer.write("\n")



from contextlib import ExitStack
from tabbyset.entities import TestCase
from tabbyset.file_formats.abc.abstract_test_cases_writer import ITestCasesWriter


class MultiTestCaseWriter(ITestCasesWriter):
    """
    A writer that writes test cases to multiple files at ones.

    :param test_case_writers: The writers to write the test cases to.
    """
    def __init__(self, *test_case_writers: ITestCasesWriter):
        self.writers = list(test_case_writers)

    def write(self, test_case: TestCase):
        for writer in self.writers:
            writer.write(test_case)

    def write_many(self, test_cases):
        for writer in self.writers:
            writer.write_many(test_cases)

    def close(self):
        exit_stack = ExitStack()
        for writer in self.writers:
            exit_stack.push(writer)
        exit_stack.close()

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""This file is fully generated by ChatGPT"""
from datetime import timedelta, date


class DateRange:
    def __init__(self, start_date: date, end_date: date):
        """
        Initialize a DateRange object.

        Parameters:
        - start_date (date): The start date of the range.
        - end_date (date): The end date of the range.

        Raises:
        - ValueError: If start_date is after end_date.
        """
        self.start_date: date = start_date
        self.end_date: date = end_date

        if start_date > end_date:
            raise ValueError("Start date must be on or before end date")

    def __repr__(self) -> str:
        """
        Return a string representation of the DateRange object.
        """
        return f"DateRange({self.start_date}, {self.end_date})"

    def __str__(self) -> str:
        """
        Return a human-readable string representation of the DateRange object.
        """
        return f"{self.start_date} to {self.end_date}"

    def __contains__(self, other_date: date) -> bool:
        """
        Check if a given date is within the date range.

        Parameters:
        - other_date (date): The date to check.

        Returns:
        - bool: True if other_date is within the date range, False otherwise.
        """
        return self.start_date <= other_date <= self.end_date

    def __iter__(self):
        """
        Iterate over the dates in the date range.
        """
        current_date = self.start_date
        while current_date <= self.end_date:
            yield current_date
            current_date += timedelta(days=1)

    def length(self) -> int:
        """
        Calculate the length of the date range in days.

        Returns:
        - int: The length of the date range.
        """
        return (self.end_date - self.start_date).days + 1

    def overlaps(self, other_range: 'DateRange') -> bool:
        """
        Check if the current date range overlaps with another date range.

        Parameters:
        - other_range (DateRange): The other date range to check.

        Returns:
        - bool: True if the date ranges overlap, False otherwise.
        """
        return (
                self.start_date <= other_range.end_date
                and self.end_date >= other_range.start_date
        )

    def merge(self, other_range: 'DateRange') -> 'DateRange':
        """
        Merge the current date range with another overlapping date range.

        Parameters:
        - other_range (DateRange): The other date range to merge.

        Returns:
        - DateRange: A new DateRange representing the merged date range.

        Raises:
        - ValueError: If the date ranges do not overlap.
        """
        if not self.overlaps(other_range):
            raise ValueError("Ranges do not overlap and cannot be merged")

        new_start_date = min(self.start_date, other_range.start_date)
        new_end_date = max(self.end_date, other_range.end_date)

        return DateRange(new_start_date, new_end_date)

    @classmethod
    def union(cls, range1: 'DateRange', range2: 'DateRange') -> 'DateRange':
        """
        Create a new DateRange from the union of two overlapping date ranges.

        Parameters:
        - range1 (DateRange): The first date range.
        - range2 (DateRange): The second date range.

        Returns:
        - DateRange: A new DateRange representing the union of the two date ranges.

        Raises:
        - ValueError: If the date ranges do not overlap.
        """
        if not range1.overlaps(range2):
            raise ValueError("Ranges do not overlap and cannot be unioned")

        new_start_date = min(range1.start_date, range2.start_date)
        new_end_date = max(range1.end_date, range2.end_date)

        return cls(new_start_date, new_end_date)



from dataclasses import dataclass
from typing import Callable, Optional

Categorizer = Callable[[dict], str]


@dataclass(frozen=True)
class MultiheaderConfig:
    row_category_prefix: str
    header_category_postfix: str
    categorizer: Categorizer

    column_before_row_category: Optional[str] = None



import sys
from tqdm import tqdm
from tabbyset.file_formats.csv1.reader import Csv1Reader
from tabbyset.file_formats.csv1.writer import Csv1Writer
from .folder import PathParam, Folder

def chunkify_csv1_file(input_file: PathParam, chunks_folder: PathParam, chunk_size: int = 3000) -> None:
    chunks_folder = Folder(chunks_folder)
    progress_bar = tqdm(unit='chunks', file=sys.stdout)
    with Csv1Reader(input_file) as reader:
        chunks_counter = 0
        written_in_chunk = 0
        current_writer = None
        for test_case in reader:
            if written_in_chunk == 0:
                chunks_counter += 1
                current_writer = Csv1Writer(chunks_folder.get_file_path(f'chunk_{chunks_counter}.csv'))
            current_writer.write(test_case)
            written_in_chunk += 1
            progress_bar.set_postfix_str(f'{written_in_chunk} test cases in chunk {chunks_counter}')
            if written_in_chunk >= chunk_size:
                progress_bar.update(1)
                written_in_chunk = 0
                current_writer.close()


import hashlib
import pickle

def dhash(input_data) -> int:
    """
    Calculates deterministic hash
    """
    # Serialize the input data to bytes
    serialized_data = pickle.dumps(input_data)

    # Create a new sha256 hash object
    hash_object = hashlib.sha256()

    # Update the hash object with the serialized data
    hash_object.update(serialized_data)

    # Get the hexadecimal representation of the hash
    hash_hex = hash_object.hexdigest()

    # Truncate the hash to 60 bits (15 hex digits)
    truncated_hash_hex = hash_hex[:15]

    # Convert the truncated hexadecimal hash to an integer
    hash_int = int(truncated_hash_hex, 16)

    return hash_int


from .flex_table.constants import EMPTY_VALUE
from .flex_table.table_queries import (
    Equal as equal,
    NotEqual as not_equal,
    GreaterThan as greater_than,
    GreaterThanOrEqual as greater_than_or_equal,
    LessThan as less_than,
    LessThanOrEqual as less_than_or_equal
)

not_empty = not_equal(EMPTY_VALUE)


from __future__ import annotations
import copy

from ..utils.dhash import dhash
from ..utils.flex_table import FlexTable, FlexTableRow
from typing import List, Union, Optional


class TestCase:
    """
    Class representing an Exactpro internal test case.

    :param name: The name of the test case.
    :param description: The description of the test case (Supported only in CSV1 and is not recommended to use in business logic).
    :param steps: The steps of the test case in the form of table.
    """
    name: str
    description: str
    id: Optional[str]

    _steps: FlexTable

    def __init__(self,
                 name: str,
                 steps: Union[List[FlexTableRow], FlexTable],
                 description: str = '',
                 id: Optional[str] = None):
        self.name = name
        self.set_steps(steps)
        self.description = description
        self.id = id

    @property
    def steps(self) -> FlexTable:
        """
        The steps of the test case in the form of table.  See more in `FlexTable`.
        """
        return self._steps

    @steps.setter
    def steps(self, steps: Union[List[FlexTableRow], FlexTable]) -> None:
        """
        Rewrites the steps of the test case.
        """
        self.set_steps(steps)

    def set_steps(self, steps: Union[List[FlexTableRow], FlexTable]) -> None:
        """
        Rewrites the steps of the test case.
        """
        if not isinstance(steps, FlexTable):
            steps = FlexTable(steps)
        self._steps = steps

    def __eq__(self, other: TestCase) -> bool:
        return self.name == other.name and self.steps == other.steps

    def __repr__(self):
        return f"TestCase(name={repr(self.name)}, steps={self.steps}, id={repr(self.id)})"

    def __copy__(self) -> TestCase:
        return TestCase(name=self.name,
                        steps=self.steps.copy(),
                        description=self.description,
                        id=self.id)

    def __deepcopy__(self, memo) -> TestCase:
        return TestCase(name=self.name,
                        steps=self.steps.copy(deep=True),
                        description=self.description,
                        id=self.id)

    def copy(self) -> TestCase:
        """
        :return: A copy of the test case.
        """
        new_tc = copy.deepcopy(self)
        return new_tc

    def __hash__(self) -> int:
        return dhash(('TestCase', hash(self.steps)))



def annotation_traces_vs_test_case(venue: str, folder_traces,  # pylint: disable=too-many-arguments
                                   folder_cases,
                                   coverage_script: str,
                                   folder_output,
                                   include_code_coverage: bool = False,
                                   save_details: bool = False,
                                   exclude_annotations: list = None):
    """
    This adapter runs `tabbyset.scripts.run_select_subset_sqlite`
    with the same interface as in `ex_subset_selector.main.annotation_traces_vs_test_case`.

    These parameters are ignored, as they are not used in the `run_select_subset_sqlite` function:
    - exclude_annotations
    """
    raise ModuleNotFoundError('Select subset SQLite was moved to the proprietary module since TabbySet went Open Source. '
                              'Use "from ex_subset_selector.select_subset_sqlite import annotation_traces_vs_test_case".')



from .csv1 import *
from .csv2 import *
from .exceptions import *
from .common import MultiheaderConfig, MultiheaderCategorizer, set_default_multiheader_config, FileParsingLogger
from .tc_jsonl import *
from .multiheader_csv import *
from .tcs_jsonl import *
from .glob_patterns import GlobPatterns



from tabbyset.entities.test_case import TestCase

def tc_to_dict(tc: TestCase) -> dict:
    return {
        "name": tc.name,
        "description": tc.description,
        "id": tc.id,
        "steps": list(tc.steps),
    }

def dict_to_tc(tc_dict: dict) -> TestCase:
    return TestCase(
        name=tc_dict["name"],
        steps=tc_dict["steps"],
        description=tc_dict.get("description", ""),
        id=tc_dict.get("id", None)
    )


from abc import ABC
from typing import Optional

from tabbyset.file_formats.abc.abstract_test_cases_reader import AbstractTestCasesReader as ATCRBase
from tabbyset.entities.test_case import TestCase


class AbstractTestCasesReader(ATCRBase, ABC):
    @staticmethod
    def _postprocess_test_case(test_case: TestCase, instrument: Optional[str]) -> TestCase:
        if '' in test_case.steps.columns:
            test_case.steps.remove_column('')
        if 'Symbol' not in test_case.steps.columns:
            test_case.steps['Symbol'] = instrument or ''
        return test_case


from typing import Type
from .test_case import TestCaseAssertions
from tabbyset.file_formats.abc.abstract_test_cases_reader import AbstractTestCasesReader
from tabbyset.file_formats import Csv1Reader, Csv2Reader, GlobPatterns
from tabbyset.utils import Folder, PathParam


class ExGeneratorTesting(TestCaseAssertions):
    """
    Class containing unit tests for the generated CSV files.
    """

    def compare_csv1_folders(self, folder1: Folder, folder2: Folder):
        """
        Compare two folders containing CSV1 files.
        """
        csv1_pattern = GlobPatterns.csv1_pattern(deep=True)
        for file1, file2 in zip(folder1.glob(csv1_pattern), folder2.glob(csv1_pattern)):
            with self.subTest(test_suite=file1.name.replace('.csv', '')):
                self.compare_custom_files(file1, file2, reader_class=Csv1Reader)

    def compare_csv2_folders(self, folder1: Folder, folder2: Folder):
        """
        Compare two folders containing CSV2 files.
        """
        csv2_pattern = GlobPatterns.csv2_pattern(deep=True)
        for file1, file2 in zip(folder1.glob(csv2_pattern), folder2.glob(csv2_pattern)):
            with self.subTest(test_suite=file1.name.replace('.matrix.csv', '')):
                self.compare_custom_files(file1, file2, reader_class=Csv2Reader)

    def compare_csv2_multiheader_folders(self, folder1: Folder, folder2: Folder):
        """
        Compare two folders containing CSV2 files with multiheaders.
        """
        csv2_pattern = GlobPatterns.csv2_pattern(deep=True)
        for file1, file2 in zip(folder1.glob(csv2_pattern), folder2.glob(csv2_pattern)):
            with self.subTest(test_suite=file1.name.replace('.matrix.csv', '')):
                self.compare_custom_files(file1, file2, reader_class=Csv2Reader, is_multiheaders=True)

    def compare_custom_files(self,
                             csv1_file1: PathParam,
                             csv1_file2: PathParam,
                             reader_class: Type[AbstractTestCasesReader] = Csv1Reader,
                             is_multiheaders: bool = False):
        if isinstance(reader_class, Csv2Reader) and is_multiheaders:
            reader1 = reader_class(csv1_file1, multiheader=True)
            reader2 = reader_class(csv1_file2, multiheader=True)
        else:
            reader1 = reader_class(csv1_file1)
            reader2 = reader_class(csv1_file2)
        for testcase1, testcase2 in zip(reader1, reader2):
            with self.subTest(test_case=testcase1.name):
                self.assertTestCasesEqual(testcase1, testcase2)
        reader1.close()
        reader2.close()



import copy
from typing import Optional, List, Union, TextIO
from itertools import zip_longest
from ..abc import AbstractTestCasesReader
from ..common import split_row, complete_row
from ..exceptions import FileParsingException
from tabbyset.file_formats.common.multiheader_csv import MultiheaderConfig
from tabbyset.file_formats.common.multiheader_csv.core import MultiheaderCsvCore
from tabbyset.utils.flex_table.constants import EMPTY_VALUE
from tabbyset.utils.warnings import libwarn
from tabbyset.utils.folder import PathParam
from tabbyset.file_formats.constants import TEST_CASE_END_LABEL, TEST_CASE_START_LABEL
from tabbyset.entities.test_case import TestCase


class Csv2Reader(AbstractTestCasesReader):
    """
    A reader for the CSV2 format.

    Csv2Reader is iterable. All Python iteration methods are supported.
    In order  to iterate over all test cases, you also can use `for` loop:

    >>> reader = Csv2Reader('path/to/file.csv')
    ... for test_case in reader:
    ...     print(test_case)

    :param file: The path of the file or an open file object.
    :param multiheader: The flag to specify explicitly if you want use multiheader or not. Default is None, what means that the reader will decide automatically.
    """
    _multiheader: Optional[bool] = None
    _multiheader_core: MultiheaderCsvCore
    _global_columns: Optional[Union[list[str], dict[str, list[str]]]] = None

    def __init__(self,
                 file: Union[PathParam, TextIO],
                 *_,
                 multiheader: Optional[bool] = None,
                 multiheader_config: Optional[MultiheaderConfig] = None):
        super().__init__(file)
        self._multiheader = multiheader
        if multiheader_config:
            if self._multiheader is False:
                libwarn('The multiheader_config parameter is ignored because the multiheader is not used.',
                        category=RuntimeWarning)
        self._multiheader_core = MultiheaderCsvCore(multiheader_config)

    @property
    def global_columns(self) -> Union[list[str], dict[str, list[str]]]:
        if self._global_columns is None:
            next(self)
            self.restart_reading()
        if self._multiheader:
            result = copy.deepcopy(self._global_columns)
            for key in result:
                result[key].remove('')
            return result
        return self._global_columns


    @classmethod
    def get_default_multiheader_config(cls) -> MultiheaderConfig:
        return MultiheaderCsvCore.config

    @classmethod
    def set_default_multiheader_config(cls, multiheader_config: MultiheaderConfig):
        MultiheaderCsvCore.config = multiheader_config

    def _parse_as_text(self):
        # TODO: Add id parsing
        self._multiheader_core.reset_headers()
        first_column_index = 0

        csvreader = self._prepare_csv_reader()

        united_columns: Optional[List[str]] = None
        columns_definitions_read = False

        current_tc_started = False
        current_tc_ended = False
        current_tc_columns: Optional[List[str]] = None
        current_tc_name: Optional[str] = None
        current_tc_content: List[dict] = []

        line_number = 0

        def create_exception(message: str) -> FileParsingException:
            return self._create_reader_exception(message, line_number)

        for row in csvreader:
            line_number += 1

            row = self._strip_row_right(row)
            row_length = len(row)
            first_row_item = row[first_column_index] if row_length > first_column_index else ''

            if row_length == 0 or all('' == s for s in row):
                continue

            if not columns_definitions_read and first_row_item != TEST_CASE_START_LABEL:
                read_line_multiheader_result = self._multiheader_core.read_line(row)
                is_multiheader_columns = read_line_multiheader_result.is_header
                if self._multiheader is None:
                    self._multiheader = is_multiheader_columns
                if self._multiheader != is_multiheader_columns:
                    if self._multiheader:
                        raise create_exception('Multiheader columns are not defined')
                    else:
                        raise create_exception('Multiheader columns are defined, but multiheader is not enabled')

                if is_multiheader_columns:
                    if read_line_multiheader_result.error_msg:
                        raise create_exception(read_line_multiheader_result.error_msg)
                else:
                    if united_columns is None:
                        united_columns = row
                continue

            # Lookup for name
            if first_row_item == TEST_CASE_START_LABEL:
                # Check if columns are defined
                if not columns_definitions_read:
                    if self._multiheader:
                        if not self._multiheader_core.headers:
                            raise create_exception('Multiheader columns are not defined')
                    else:
                        if united_columns is None:
                            raise create_exception('Columns are not defined')
                    columns_definitions_read = True
                    if self._multiheader:
                        self._global_columns = self._multiheader_core.headers
                    else:
                        self._global_columns = united_columns

                if current_tc_started:
                    raise create_exception('Started test case is started again')
                if row_length <= 1:
                    raise create_exception('Start case row should have at least 2 items')
                current_tc_name = row[first_column_index + 1]
                current_tc_ended = False
                current_tc_started = True
                if not current_tc_name:
                    raise create_exception('Test case name not found')
                continue

            # Add content
            if first_row_item not in [TEST_CASE_START_LABEL, TEST_CASE_END_LABEL] and current_tc_started:
                # Checking table header
                # Multiheader files have no local headers
                if current_tc_columns is None and not self._multiheader:
                    if row_length > len(united_columns):
                        # Trim row if it has more items than columns
                        columns_part, extra_part = split_row(row, len(united_columns))
                        current_tc_columns = columns_part
                        current_tc_content.append(columns_part)
                        # Last part
                        if any(extra_part):
                            raise create_exception(
                                f'Extra test case columns - {extra_part}')
                        continue
                    if row_length < len(united_columns):
                        current_tc_columns = complete_row(row, len(united_columns))
                        continue
                    current_tc_columns = row
                    continue

                if self._multiheader:
                    read_line_multiheader_result = self._multiheader_core.read_line(row)
                    if read_line_multiheader_result.error_msg:
                        raise create_exception(read_line_multiheader_result.error_msg)
                    current_row_columns = read_line_multiheader_result.columns
                else:
                    if current_tc_columns is None:
                        raise create_exception('Columns are not defined for test case')
                    current_row_columns = current_tc_columns

                # Checking table row
                if row_length > len(current_row_columns):
                    # Trim row if it has more items than columns
                    columns_part, extra_part = split_row(row, len(current_row_columns))
                    current_row_as_dict = dict(zip_longest(current_row_columns, columns_part, fillvalue=EMPTY_VALUE))
                else:
                    current_row_as_dict = dict(zip_longest(current_row_columns, row, fillvalue=EMPTY_VALUE))
                if self._multiheader:
                    category_check_result = self._multiheader_core.check_row_category(current_row_as_dict,
                                                                                      read_line_multiheader_result.category)
                    if not category_check_result[0]:
                        raise create_exception(category_check_result[1])
                current_tc_content.append(current_row_as_dict)
                continue

            # Add new testcase and reset data
            if first_row_item == TEST_CASE_END_LABEL:
                if not current_tc_started:
                    raise create_exception('Not started case tries to end')

                new_test_case = self._postprocess_test_case(TestCase(name=current_tc_name, steps=current_tc_content))
                yield new_test_case

                current_tc_columns = None
                current_tc_started = False
                current_tc_ended = True
                current_tc_name = None
                current_tc_content = []
        if current_tc_started and not current_tc_ended:
            raise create_exception('Last test case is not closed')



from abc import ABC, abstractmethod
from decimal import Decimal, InvalidOperation
from typing import Union, Any
from .constants import EMPTY_VALUE

from .typing import FlexTableValue


class QueryStatement(ABC):
    _base_value: FlexTableValue

    def __init__(self, base_value: FlexTableValue):
        self._base_value = self._try_number_cast(base_value)

    def apply(self, v: FlexTableValue) -> bool:
        return self._apply_to_casted(self._try_number_cast(v), self._base_value)

    @classmethod
    @abstractmethod
    def _apply_to_casted(cls, v1: FlexTableValue, v2: FlexTableValue) -> bool:
        pass  # pragma: no cover

    @staticmethod
    def _try_number_cast(v: FlexTableValue) -> FlexTableValue:
        try:
            return Decimal(v)
        except InvalidOperation:
            return v


class Equal(QueryStatement):
    @classmethod
    def _apply_to_casted(cls, v1: FlexTableValue, v2: FlexTableValue) -> bool:
        return v1 == v2


class NotEqual(QueryStatement):
    @classmethod
    def _apply_to_casted(cls, v1: FlexTableValue, v2: FlexTableValue) -> bool:
        return v1 != v2


class GreaterThan(QueryStatement):
    @classmethod
    def _apply_to_casted(cls, v1: FlexTableValue, v2: FlexTableValue) -> bool:
        return v1 > v2


class GreaterThanOrEqual(QueryStatement):
    @classmethod
    def _apply_to_casted(cls, v1: FlexTableValue, v2: FlexTableValue) -> bool:
        return v1 >= v2


class LessThan(QueryStatement):
    @classmethod
    def _apply_to_casted(cls, v1: FlexTableValue, v2: FlexTableValue) -> bool:
        return v1 < v2


class LessThanOrEqual(QueryStatement):
    @classmethod
    def _apply_to_casted(cls, v1: FlexTableValue, v2: FlexTableValue) -> bool:
        return v1 <= v2


ParsableQueryStatement = Union[FlexTableValue, QueryStatement]
DictQuery = dict[str, ParsableQueryStatement]


def parse_query_statement(query: ParsableQueryStatement) -> QueryStatement:
    if isinstance(query, QueryStatement):
        return query
    if not isinstance(query, str):
        return Equal(query)
    if query.startswith('= '):
        return Equal(query[2:])
    if query.startswith('!= '):
        return NotEqual(query[3:])
    if query.startswith('> '):
        return GreaterThan(query[2:])
    if query.startswith('>= '):
        return GreaterThanOrEqual(query[3:])
    if query.startswith('< '):
        return LessThan(query[2:])
    if query.startswith('<= '):
        return LessThanOrEqual(query[3:])
    return Equal(query)


def parse_dict_query(query_dict: DictQuery) -> dict[str, QueryStatement]:
    return {k: parse_query_statement(v) for k, v in query_dict.items()}


def apply_query_to_dict(query_dict: dict[str, QueryStatement], value_dict: dict[str, FlexTableValue]) -> bool:
    return all(q.apply(value_dict.get(k, EMPTY_VALUE)) for k, q in query_dict.items())



import difflib
from decimal import Decimal, InvalidOperation
from tabbyset.utils.flex_table import FlexTable, dict_row_to_list
from typing import Union
from .diff import TableCellDiff, ColoredString, ConsoleColor


def _dict_row_to_formatted_list(row: dict, columns: list[str]) -> list[str]:
    result = []
    for v in dict_row_to_list(row, columns):
        try:
            result.append(str(Decimal(v)))
        except InvalidOperation:
            result.append(str(v))
    return result


def _get_columns_diff_sequence_matcher(expected: FlexTable, actual: FlexTable):
    expected_columns = expected.columns
    actual_columns = actual.columns
    matcher = difflib.SequenceMatcher(None, expected_columns, actual_columns)
    opcodes = matcher.get_opcodes()
    columns_diff = []
    merged_columns = []
    for tag, expected_index_start, expected_index_end, actual_index_start, actual_index_end in opcodes:
        expected_indexes = range(expected_index_start, expected_index_end)
        actual_indexes = range(actual_index_start, actual_index_end)
        if tag == "equal":
            columns_diff += expected_columns[expected_index_start:expected_index_end]
            merged_columns += expected_columns[expected_index_start:expected_index_end]
        elif tag == "replace":
            for expected_index in expected_indexes:
                columns_diff.append(TableCellDiff(added_value=expected_columns[expected_index]))
                merged_columns.append(expected_columns[expected_index])
            for actual_index in actual_indexes:
                columns_diff.append(TableCellDiff(removed_value=actual_columns[actual_index]))
                merged_columns.append(actual_columns[actual_index])
        elif tag == "insert":
            columns_diff += [TableCellDiff(removed_value=actual_columns[j]) for j in actual_indexes]
            merged_columns += [actual_columns[j] for j in actual_indexes]
        elif tag == "delete":
            columns_diff += [TableCellDiff(added_value=expected_columns[i]) for i in expected_indexes]
            merged_columns += [expected_columns[i] for i in expected_indexes]
    return columns_diff, merged_columns


def _get_columns_diff_set(expected: FlexTable, actual: FlexTable):
    expected_columns = set(expected.columns)
    actual_columns = set(actual.columns)
    columns_diff = [TableCellDiff(added_value=column) for column in expected_columns - actual_columns]
    columns_diff += [TableCellDiff(removed_value=column) for column in actual_columns - expected_columns]
    merged_columns = list(expected_columns | actual_columns)
    return columns_diff, merged_columns


def get_columns_diff(table1: FlexTable, table2: FlexTable):
    expected_columns = set(table1.columns)
    actual_columns = set(table2.columns)
    if expected_columns == actual_columns:
        return list(table1.columns), list(table2.columns)
    # Logic for diff is swapped, so it is fixed here
    columns_diff, merged_columns = _get_columns_diff_sequence_matcher(table2, table1)
    if len(merged_columns) > len(expected_columns | actual_columns):
        # Logic for diff is swapped, so it is fixed here
        columns_diff, merged_columns = _get_columns_diff_set(table2, table1)
    return columns_diff, merged_columns


def get_flex_tables_diff(table1: FlexTable, table2: FlexTable, merged_columns: list[str]):
    # Logic for diff is swapped, so it is fixed here
    expected_formatted = [_dict_row_to_formatted_list(row, merged_columns) for row in table2]
    actual_formatted = [_dict_row_to_formatted_list(row, merged_columns) for row in table1]
    old_messages = [str(row) for row in expected_formatted]
    new_messages = [str(row) for row in actual_formatted]
    # Create a SequenceMatcher object
    matcher = difflib.SequenceMatcher(None, old_messages, new_messages)

    # Get the list of operations that transform the old sequence into the new sequence
    opcodes = matcher.get_opcodes()

    table_diff: list[list[Union[TableCellDiff, str, ColoredString]]] = []

    # Iterate over the opcodes
    for tag, expected_index_start, expected_index_end, actual_index_start, actual_index_end in opcodes:
        expected_indexes = range(expected_index_start, expected_index_end)
        actual_indexes = range(actual_index_start, actual_index_end)
        if tag == "equal":
            # The messages are equal, so add a mapping from the old message index to the new message index
            if len(expected_indexes) < 3:
                for expected_index, actual_index in zip(expected_indexes, actual_indexes):
                    table_diff.append(list(expected_formatted[expected_index]))
            else:
                table_diff.append(list(expected_formatted[expected_indexes[0]]))
                table_diff.append([ColoredString("...", ConsoleColor.GRAY) for _ in range(len(merged_columns))])
                table_diff.append(list(expected_formatted[expected_indexes[-1]]))
        elif tag == "replace":
            # The messages are different, so add a mapping from the old message index to the new message index
            for expected_index, actual_index in zip(expected_indexes, actual_indexes):
                table_diff.append([
                    TableCellDiff(
                        removed_value=actual_formatted[actual_index][k],
                        added_value=expected_formatted[expected_index][k]
                    )
                    if actual_formatted[actual_index][k] != expected_formatted[expected_index][k]
                    else expected_formatted[expected_index][k]
                    for k in range(len(merged_columns))
                ])
            if len(expected_indexes) > len(actual_indexes):
                for expected_index_diff in range(len(expected_indexes) - len(actual_indexes)):
                    actual_row_index = actual_index_start + len(actual_indexes) + expected_index_diff
                    if actual_row_index < len(actual_formatted):
                        table_diff.append([TableCellDiff(added_value=actual_formatted[actual_row_index][k])
                                           for k in range(len(merged_columns))])
                    else:
                        expected_row_index = expected_indexes[-expected_index_diff - 1]
                        table_diff.append([TableCellDiff(removed_value=expected_formatted[expected_row_index][k])
                                           for k in range(len(merged_columns))])
            elif len(actual_indexes) > len(expected_indexes):
                for actual_index_diff in range(len(actual_indexes) - len(expected_indexes)):
                    expected_row_index = expected_index_start + len(expected_indexes) + actual_index_diff
                    if expected_row_index < len(expected_formatted):
                        table_diff.append([TableCellDiff(removed_value=expected_formatted[expected_row_index][k])
                                           for k in range(len(merged_columns))])
                    else:
                        actual_row_index = actual_indexes[-actual_index_diff - 1]
                        table_diff.append([TableCellDiff(added_value=actual_formatted[actual_row_index][k])
                                           for k in range(len(merged_columns))])
        elif tag == "insert":
            # The new message is not in the old sequence, so add a mapping from None to the new message index
            for actual_index in actual_indexes:
                table_diff.append(
                    [TableCellDiff(
                        removed_value=actual_formatted[actual_index][k]
                    ) for k in range(len(merged_columns))])
        elif tag == "delete":
            # The old message is not in the new sequence, so add a mapping from the old message index to None
            for expected_index in expected_indexes:
                table_diff.append(
                    [TableCellDiff(
                        added_value=expected_formatted[expected_index][k]
                    ) for k in range(len(merged_columns))])
    return table_diff



"""
Module containing a SQLite-based implementation of the select subset pipeline.
"""
from .run_select_subset_sqlite import run_select_subset_sqlite
from .run_adapters import annotation_traces_vs_test_case



"""
File system utilities
"""
import os.path
from typing import Literal
from tabbyset.file_formats import Csv1Reader, Csv2Reader
from .folder import Folder, PathParam


FileFormat = Literal['csv1', 'csv2']

def walk_tests_folder(folder_path: PathParam, file_format: FileFormat = 'csv1',
                      *,
                      deep: bool = False):
    """
    Walk through the test cases in a folder.
    :param folder_path: The path of the folder.
    :param file_format: The format of the test cases files.
    :param deep: If True, walk through all subfolders.
    :return: An iterator of tuples with a filepath relative to base folder and its testcases.
    """
    if not os.path.exists(folder_path):
        raise FileNotFoundError(f"Folder {folder_path} does not exist")
    folder = Folder(folder_path)
    if file_format == 'csv1':
        file_pattern = '*.csv'
    elif file_format == 'csv2':
        file_pattern = '*.matrix.csv'
    else:
        raise ValueError(f"Unknown file format {file_format}")
    if deep:
        file_pattern = f'**/{file_pattern}'
    for file in folder.glob(file_pattern):
        rel_path = os.path.relpath(file, folder.path)
        if file_format == 'csv1':
            # CSV1 files have .csv extension, however, we want to skip the matrix files
            # Glob patterns support syntax for excluding files, but its implementation is bugged, so we have to do it manually
            if any(rel_path.endswith(ext) for ext in [".matrix.csv", ".matrix.d.csv", ".matrix.expected.csv",
                                                      ".Input.csv", ".Trace.csv"]):
                continue
            with Csv1Reader(file) as reader:
                for tc in reader:
                    yield rel_path, tc
        elif file_format == 'csv2':
            with Csv2Reader(file) as reader:
                for tc in reader:
                    yield rel_path, tc